/*
 * VIAGRA Internet Relay Chat daemon, include/config.h
 * Copyright (C) 1990 Jarkko Oikarinen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef	__config_include__
#define	__config_include__

#include "setup.h"
#include "s_dconf.h"

#define OPTIONS_H

/* If you are using OPTIONS_H, you can ignore everything from here until i tell you otherwise. If not, go nuts. */
#ifdef OPTIONS_H
#include "options.h"
#endif

/* Are we compiling under cygwin? If so define CYGWIN */
/* to do so uncomment the #define and comment out #undef */
/* #define CYGWIN */
#undef CYGWIN

#ifdef CYGWIN
#define FD_SETSIZE 1024
#endif

#ifndef OPTIONS_H

/**** MAKE SURE THIS IS CORRECT ****/
/* ONLY EDIT "HARD_FDLIMIT_" and "INIT_MAXCLIENTS" */
#define HARD_FDLIMIT_ 1024
#define INIT_MAXCLIENTS 900

/* This is how many 'buffer connections' we allow... Remember, MAX_BUFFER + MAX_CLIENTS can't exceed HARD_FDLIMIT :) */
#define MAX_BUFFER 124

/* Also define this for SunOS 4.1.4_DBE */
#undef SUNDBE
#ifdef SUNDBE
#define FD_SETSIZE HARD_FDLIMIT_
#define NOFILE HARD_FDLIMIT_
#endif /* SUNDBE */

/* DPATH SPATH CPATH MPATH KPATH - directoy and files locations Full
 * pathnames and defaults of irc system's support files. Please note
 * that these are only the recommended names and paths. Change as
 * needed. You must define these to something, even if you don't really
 * want them.
 *
 * DPATH = directory, SPATH = server executable, CPATH = conf file, MPATH
 * = MOTD KPATH = kline conf file
 *
 * For /restart to work, SPATH needs to be a full pathname (unless "." is
 * in your exec path). -Rodder Leave KPATH undefined if you want klines
 * in main conf file. HPATH is the opers help file, seen by opers on
 * /quote help.
 *
 * -Dianora */
#define DPATH "/usr/home/ircd/ircd/"
#define SPATH "/usr/home/ircd/ircd/ircd"

/* SERVICESHUB - Provides some filtering of data not needed by services */
#undef SERVICESHUB /* broken atm */

/* MAXSENDQLENGTH - Max amount of internal send buffering Max amount of
 * internal send buffering when socket is stuck (bytes) */
#define MAXSENDQLENGTH 5050000	/* Recommended value: 5050000 for EFnet */

/* BUFFERPOOL - the maximum size of the total of all sendq's.
 * Recommended value is four times MAXSENDQLENGTH. */
#define	BUFFERPOOL (4 * MAXSENDQLENGTH)

/* comstud and I have both noted that solaris 2.5 at least, takes a
 * hissy fit if you don't read a fd that becomes ready right away.
 * Unfortunately the dog3 priority code relies upon not having to read
 * a ready fd right away. If you have HTM mode set low as it is
 * normally, the server will eventually grind to a halt. Personally, I
 * think the server is faster without some of the CPU expensive
 * manipulation some of the priority code does. Your choice. but it has
 * to be defined for SOLARIS Try it and see. compare. Don't complain if
 * Solaris lags if you don't define this. I warned you. -Dianora */
#ifdef OS_SOLARIS
#define NO_PRIORITY
#else
#undef NO_PRIORITY
#endif

#endif /** OPTIONS_H **/
/******  END OF AUTOMATIC CONFIG SCRIPT COVERAGE  ******/

/* Secure Config File Path
 * Define to make the server set the specified permission upon CPATH (ircd.conf) upon
 * bootup. We recommend you enable this, and leave the permission setting at default (0600). */
#define SECURE_CPATH
#define SECURE_CPATH_PERM 0600

/* Secure Config File Path
 * Define to make the server set the specified permission upon ZCONF (ircd.ini) upon
 * bootup. We recommend you enable this, and leave the permission setting at default (0600). */
#define SECURE_DCONF
#define SECURE_DCONF_PERM 0600

/* OS Depentant ifdefs */
#ifdef SUNDBE
#define FD_SETSIZE HARD_FDLIMIT_
#define NOFILE HARD_FDLIMIT_
#endif /* SUNDBE */

#ifdef OS_SOLARIS
#define NO_PRIORITY
#else
#undef NO_PRIORITY
#endif

/* File names */
#define	CPATH "conf/ircd.conf"
#define DCONF "conf/ircd.ini"			/* Dynamic Configuration File */
#define KPATH "conf/kline.conf"
#define RPATH "conf/ircd.rules"			/* Rules File */
#define	MPATH "conf/ircd.motd"
#define OMPATH "conf/ircd.omotd"		/* Operator Message Of The Day */
#define	SMPATH "conf/ircd.smotd"

/* path to ircd log files */
#define	SYSTEMPATH "logs/system"
#define OPERPATH "logs/oper"
/* This is the number of days to keep the logs */
#define KEEPLOGS 5

/* This is the number of users that must be in a channel before 
   it will be added to the channel listing */
#define MINCHANUSERS 1

#define	PPATH "ircd.pid"
#define HPATH "conf/ircd.ohelp"
#define VPATH "conf/vhost.conf"
#define IRCD_RESTRICT "conf/channels.conf"	/* Channel Restrict file */
#define IRCD_DCCDENY "conf/dccdeny.conf"	/* DCC Restrict file */

/* Services Definitions */
#define STATS_NAME "stats.dal.net"
#define CHANSERV "ChanServ"
#define NICKSERV "NickServ"
#define MEMOSERV "MemoServ"
#define BOTSERV "BotServ"
#define OPERSERV "OperServ"
#define STATSERV "StatServ"
#define HOSTSERV "HostServ"
#define ROOTSERV "RootServ"
#define ADMINSERV "AdminServ"
#define MASSSERV "MassServ"
#define GLOBALSERV "GlobalServ"
#define AGENT "Agent"
#define X "X"
#define W "W"


#define NICKSERVATSERVICES NICKSERV "@" SERVICES_NAME
#define CHANSERVATSERVICES CHANSERV "@" SERVICES_NAME
#define MEMOSERVATSERVICES MEMOSERV "@" SERVICES_NAME
#define BOTSERVATSERVICES BOTSERV "@" SERVICES_NAME
#define OPERSERVATSTATS OPERSERV "@" SERVICES_NAME
#define STATSERVATSTATS STATSERV "@" STATS_NAME


#define HELPSERVATSTATS HOSTSERV "@" SERVICES_NAME
#define ROOTSERVATSTATS ROOTSERV "@" SERVICES_NAME

/* Don't change this... */
#define HARD_FDLIMIT (HARD_FDLIMIT_ - 10)
#define MASTER_MAX (HARD_FDLIMIT - MAX_BUFFER)

/* DONT_USE_REGISTER: this tells the compiler not to try to make
 * variables 'register', almost always a good idea on a modern
 * compiler */
#define DONT_USE_REGISTER

#include "defs.h"

/* USE_FAST_FD_ISSET - experimental! - crawl inside of FD_ISSET macro
 * This one is experimental, though it works on SUNOS and freebsd It is
 * NOT portable, and "crawls" carnally inside the FD_ISSET macro but
 * should be a faster way of doing select() in s_bsd.c
 *
 * The idea is to pre-calculate the bit map mask needed shifting it over
 * as needed, the FD_ISSET calculates the bitmask and array offset
 * every single time. */
/* Checked for by configure now. */
/* #define USE_FAST_FD_ISSET */

/* TS_MAX_DELTA and TS_WARN_DELTA -  allowed delta for TS when another
 * server connects.
 *
 * If the difference between my clock and the other server's clock is
 * greater than TS_MAX_DELTA, I send out a warning and drop the links.
 *
 * If the difference is less than TS_MAX_DELTA, I just sends out a warning
 * but don't drop the link.
 *
 * TS_MAX_DELTA currently set to 30 minutes to deal with older timedelta
 * implementation.  Once pre-hybrid5.2 servers are eradicated, we can
 * drop this down to 90 seconds or so. --Rodder */
#define TS_MAX_DELTA 5000	/* seconds */
#define TS_WARN_DELTA 30	/* seconds */

/* TS5
 * This will mean the oldest version of a channel will always remain after a
 * split, whereas TS3 behaviour was to allow any TS in an opless channel.
 *
 * This effectively will make splitting servers for ops useless, but could
 * cause desync on a mixed TS3/TS5 network.
 *
 * See doc/ts5.txt for more details
 * This is an EFnet requirement now.
 */
#define TS5

/* DEFAULT_KLINE_TIME
 *
 * Define this to the default time for a kline (in minutes) for klines with
 * unspecified times. undefine this for all klines with unspecified times
 * to be perm. (if defined, a kline with zero time will be perm). -- lucas */
#define DEFAULT_KLINE_TIME 30

/* CONNECTFAST
 *
 * if not defined, only allow 1 connection per listening socket
 * every two seconds. */
#define CONNECTFAST

/* LOCKFILE - Exclusive use of ircd.conf and kline.conf during writes
 *
 * This creates a lockfile prior to writes to ircd.conf or kline.conf, and
 * can be used in conjunction with viconf (included in the tools
 * directory). This prevents loss of data when klines are added while
 * someone is manually editting the file.  File writes will be retried
 * at the next KLINE, ZLINE, REHASH, or after CHECK_PENDING_KLINES
 * minutes have elapsed.
 *
 * If you do not wish to use this feature, leave LOCKFILE #undef */
#define LOCKFILE "/tmp/ircd.conf.lock"
#define	CHECK_PENDING_KLINES 10 /* in minutes */

/* FOLLOW_IDENT_RFC
 *
 * From RFC 1413, "The Identification Protocol is not intended as an
 * authorization or access control protocol ... The use of the
 * information returned by this protocol for other than auditing is
 * strongly discouraged."
 *
 * Defining this allows all users to set whatever username they wish,
 * despite what may be discovered by ident.  While we may get a valid
 * ident response with a different username than submitted by the
 * client, this option will cause the ident response to be discarded. */
#undef FOLLOW_IDENT_RFC

/* CLIENT_SERVER - Don't be so fascist about idle clients ;) changes
 * behaviour of HTM code to make clients lag less. */
#define CLIENT_SERVER

/*
 * IGNORE_FIRST_CHAR - define this for NO_MIXED_CASE if you wish to
 * ignore the first character
 */
#define IGNORE_FIRST_CHAR

/*
 * ZLINES_IN_KPATH - put (/quote) ZLINES in your KPATH file. Define
 * this if you want ZLINES in your KPATH file.  Otherwise ZLINES will
 * be added to the main ircd.conf
 */
#ifdef KPATH
#define ZLINES_IN_KPATH
#endif /* KPATH */

/*
 * HIGHEST_CONNECTION - track highest connection count Define this if
 * you want to keep track of your max connections.
 */
#define HIGHEST_CONNECTION

/*
 * USERNAMES_IN_TRACE - show usernames in trace Define this if you want
 * to see usernames in /trace.
 */
#define USERNAMES_IN_TRACE

/*
 * DO_IDENTD - check identd if you undefine this, ircd will never check
 * identd regardless of @'s in I:lines.  You must still use @'s in your
 * I: lines to get ircd to do ident lookup even if you define this.
 */
#define DO_IDENTD

/*
 * NON_REDUNDANT_KLINES - If you want the server to flag and not apply
 * redundant klines
 */
#define NON_REDUNDANT_KLINES

/*
 * CLIENT_COUNT - keep a running count of Local & Global users also
 * redefine the /USERS command
 */
#define CLIENT_COUNT

/* MOTD_WAIT - minimum seconds between use of MOTD, INFO, HELP, LINKS
 * before max use count is reset -Dianora  */
#define MOTD_WAIT 10

/* WHOIS_WAIT - minimum seconds between use of WHOIS
 * before max use count is reset */
/* #define WHOIS_WAIT 3 */

/* SHOW_UH - show the user@host everywhere */
#define SHOW_UH
#ifdef SHOW_UH
#define USERNAMES_IN_TRACE
#endif

/* SHOW_INVISIBLE_LUSERS - show invisible clients in LUSERS As defined
 * this will show the correct invisible count for anyone who does
 * LUSERS on your server. On a large net this doesnt mean much, but on
 * a small net it might be an advantage to undefine it. */
#define	SHOW_INVISIBLE_LUSERS

/* MAXIMUM LINKS - max links for class 0 if no Y: line configured
 *
 * This define is useful for leaf nodes and gateways. It keeps you from
 * connecting to too many places. It works by keeping you from
 * connecting to more than "n" nodes which you have C:blah::blah:6667
 * lines for.
 *
 * Note that any number of nodes can still connect to you. This only
 * limits the number that you actively reach out to connect to.
 *
 * Leaf nodes are nodes which are on the edge of the tree. If you want to
 * have a backup link, then sometimes you end up connected to both your
 * primary and backup, routing traffic between them. To prevent this,
 * #define MAXIMUM_LINKS 1 and set up both primary and secondary with
 * C:blah::blah:6667 lines. THEY SHOULD NOT TRY TO CONNECT TO YOU, YOU
 * SHOULD CONNECT TO THEM.
 *
 * Gateways such as the server which connects Australia to the US can do a
 * similar thing. Put the American nodes you want to connect to in with
 * C:blah::blah:6667 lines, and the Australian nodes with C:blah::blah
 * lines. Have the Americans put you in with C:blah::blah lines. Then
 * you will only connect to one of the Americans.
 *
 * This value is only used if you don't have server classes defined, and a
 * server is in class 0 (the default class if none is set). */
#define MAXIMUM_LINKS 1

/* M4_PREPROC - run ircd.conf through m4 preprocessor To use m4 as a
 * preprocessor on the ircd.conf file, define M4_PREPROC. The server
 * will then call m4 each time it reads the ircd.conf file, reading m4
 * output as the server's ircd.conf file. m4 support is being moved
 * into viconf.c m4 will be run on ircd.conf.m4 producing ircd.conf */
#undef M4_PREPROC

/* NOISY_HTM - should HTM be noisy by default should be YES or NO */
#define NOISY_HTM YES

/* CLIENT_FLOOD - client excess flood threshold this controls the
 * number of bytes the server will allow a client to send to the server
 * without processing before disconnecting the client for flooding it.
 * Values greater than 8000 make no difference to the server. */
#define	CLIENT_FLOOD 2560

/*
 * CMDLINE_CONFIG - allow conf-file to be specified on command line
 * NOTE: defining CMDLINE_CONFIG and installing ircd SUID or SGID is a
 * MAJOR security problem - they can use the "-f" option to read any
 * files that the 'new' access lets them.
 */
#define	CMDLINE_CONFIG

/*
 * ANTI_NICK_FLOOD - prevents nick flooding define if you want to block
 * local clients from nickflooding
 */
#define ANTI_NICK_FLOOD

/*
 * defaults allow 5 nick changes in 20 seconds
 */
#define MAX_NICK_TIME 15
#define MAX_NICK_CHANGES 5

/* NO_SERVICES_FLUD
 */
#define NO_SERVICES_FLUD
#ifdef NO_SERVICES_FLUD
# define MAX_MSG_TIME 10 /* time in seconds */
# define MAX_MSG_COUNT 8
#endif

/* NO_AWAY_FLUD
 * reallow propregation of AWAY messages, but do not allow AWAY flooding
 * I reccomend a max of 5 AWAY's in 3 Minutes
 */
#define NO_AWAY_FLUD

#ifdef NO_AWAY_FLUD
# define MAX_AWAY_TIME 180 /* time in seconds */
# define MAX_AWAY_COUNT 5
#endif

#define ANTI_SPAM_EXIT_MESSAGE
#define ANTI_SPAM_EXIT_MESSAGE_TIME 75

/* When a host tries to connect more than 2 times
 * in 3 seconds it will block them from connecting for another 15 seconds */
#undef THROTTLE
#ifdef THROTTLE
# define CTIMES 2
# define CDELAY 3
# define CEDELAY 10
#endif

/* If you want your server to save and restore the local/max client count,
 * define this here.  It's neat, non-load-producing, and makes you
 * appear more attractive to the opposite sex. --wd */
#define SAVE_MAXCLIENT_STATS

/* Define the time (in seconds) that you want to save the statistics after.
 * Don't undef this unless you have undeffed SAVE_MAXCLIENT_STATS --sed */
#define SAVE_MAXCLIENT_STATS_TIME 3600

/*
 * STOP STOP STOP STOP STOP STOP STOP STOP STOP STOP STOP STOP STOP
 *
 * You shouldn't change anything below this line, unless absolutely
 * needed.
 */

/*
 * USE_NOQUIT use the noquit system to save bandwidth and network stress on netsplits.
 */
#define USE_NOQUIT

/*
 * PING_NAZI
 *
 * be nazi-ish about pings (re-check every client connect, user registration, etc)
 */
#undef PING_NAZI

/*
 * ALWAYS_SEND_DURING_SPLIT
 * on a large network, if your server is carrying large amounts of clients,
 * and your server splits from the main network, the amount of allocated
 * dbufs will skyrocket as buffers fill up with QUIT messages. This code
 * attempts to combat this by sending out data whenever possible during a
 * split. - lucas
 */
#define ALWAYS_SEND_DURING_SPLIT

/* INITIAL_DBUFS - how many dbufs to preallocate */
#define INITIAL_DBUFS 1024 /* preallocate 4 megs of dbufs */

/*
 * MAXBUFFERS - increase socket buffers
 *
 * Increase send & receive socket buffer up to 64k, keeps clients at 8K
 * and only raises servers to 64K
 */
#define MAXBUFFERS

/*
 * PORTNUM - default port where ircd resides Port where ircd resides.
 * NOTE: This *MUST* be greater than 1024 if you plan to run ircd under
 * any other uid than root.
 */
#define PORTNUM 7000 /* 7000 for DALnet */

/*
 * MAXCONNECTIONS - don't touch - change the HARD_FDLIMIT_ instead
 * Maximum number of network connections your server will allow.  This
 * should never exceed max. number of open file descrpitors and wont
 * increase this. Should remain LOW as possible. Most sites will
 * usually have under 30 or so connections. A busy hub or server may
 * need this to be as high as 50 or 60. Making it over 100 decreases
 * any performance boost gained from it being low. if you have a lot of
 * server connections, it may be worth splitting the load over 2 or
 * more servers. 1 server = 1 connection, 1 user = 1 connection. This
 * should be at *least* 3: 1 listen port, 1 dns port + 1 client
 *
 * Change the HARD_FDLIMIT_ instead
 */
#define MAXCONNECTIONS HARD_FDLIMIT

/*
 * NICKNAMEHISTORYLENGTH - size of WHOWAS array this defines the length
 * of the nickname history.  each time a user changes nickname or signs
 * off, their old nickname is added to the top of the list. NOTE: this
 * is directly related to the amount of memory ircd will use whilst
 * resident and running - it hardly ever gets swapped to disk!  Memory
 * will be preallocated for the entire whowas array when ircd is
 * started.
 */
#define NICKNAMEHISTORYLENGTH 1024

/*
 * TIMESEC - Time interval to wait and if no messages have been
 * received, then check for PINGFREQUENCY and CONNECTFREQUENCY
 */
#define TIMESEC 5		/* Recommended value: 5 */

/*
 * PINGFREQUENCY - ping frequency for idle connections If daemon
 * doesn't receive anything from any of its links within PINGFREQUENCY
 * seconds, then the server will attempt to check for an active link
 * with a PING message. If no reply is received within (PINGFREQUENCY *
 * 2) seconds, then the connection will be closed.
 */
#define PINGFREQUENCY 120	/* Recommended value: 120 */

/*
 * CONNECTFREQUENCY - time to wait before auto-reconencting If the
 * connection to to uphost is down, then attempt to reconnect every
 * CONNECTFREQUENCY  seconds.
 */
#define CONNECTFREQUENCY 600	/* Recommended value: 600 */

/*
 * HANGONGOODLINK and HANGONGOODLINK Often net breaks for a short time
 * and it's useful to try to establishing the same connection again
 * faster than CONNECTFREQUENCY would allow. But, to keep trying on bad
 * connection, we require that connection has been open for certain
 * minimum time (HANGONGOODLINK) and we give the net few seconds to
 * steady (HANGONRETRYDELAY). This latter has to be long enough that
 * the other end of the connection has time to notice it broke too.
 * 1997/09/18 recommended values by ThemBones for modern Efnet
 */

#define HANGONRETRYDELAY 60	/* Recommended value: 30-60 seconds */
#define HANGONGOODLINK 3600	/* Recommended value: 30-60 minutes */

/*
 * WRITEWAITDELAY - Number of seconds to wait for write to complete if
 * stuck.
 */
#define WRITEWAITDELAY 15	/* Recommended value: 15 */

/*
 * CONNECTTIMEOUT - Number of seconds to wait for a connect(2) call to
 * complete. NOTE: this must be at *LEAST* 10.  When a client connects,
 * it has CONNECTTIMEOUT - 10 seconds for its host to respond to an
 * ident lookup query and for a DNS answer to be retrieved.
 */
#define	CONNECTTIMEOUT 30	/* Recommended value: 30 */

/*
 * KILLCHASETIMELIMIT - Max time from the nickname change that still
 * causes KILL automaticly to switch for the current nick of that user.
 * (seconds)
 */
#define KILLCHASETIMELIMIT 90	/* Recommended value: 90 */

/*
 * FLUD - CTCP Flood Detection and Protection
 *
 * This enables server CTCP flood detection and protection for local
 * clients. It works well against fludnets and flood clones.  The
 * effect of this code on server CPU and memory usage is minimal,
 * however you may not wish to take the risk, or be fundamentally
 * opposed to checking the contents of PRIVMSG's (though no privacy is
 * breached).  This code is not useful for routing only servers (ie,
 * HUB's with little or no local client base), and the hybrid team
 * strongly recommends that you do not use FLUD with HUB. The following
 * default thresholds may be tweaked, but these seem to work well.
 */
#define FLUD

/*
 * ANTI_SPAMBOT parameters, don't touch these if you don't understand
 * what is going on.
 *
 * if a client joins MAX_JOIN_LEAVE_COUNT channels in a row, but spends
 * less than MIN_JOIN_LEAVE_TIME seconds on each one, flag it as a
 * possible spambot. disable JOIN for it and PRIVMSG but give no
 * indication to the client that this is happening. every time it tries
 * to JOIN OPER_SPAM_COUNTDOWN times, flag all opers on local server.
 * If a client doesn't LEAVE a channel for at least 2 minutes the
 * join/leave counter is decremented each time a LEAVE is done
 *
 */
#define MIN_JOIN_LEAVE_TIME 60
#define MAX_JOIN_LEAVE_COUNT 15
#define OPER_SPAM_COUNTDOWN 5
#define JOIN_LEAVE_COUNT_EXPIRE_TIME 120

#ifdef FLUD
# define FLUD_NUM 4	/* Number of flud messages to trip alarm */
# define FLUD_TIME 3	/* Seconds in which FLUD_NUM msgs must occur */
# define FLUD_BLOCK 15	/* Seconds to block fluds */
#endif

/* ANTI_DRONE_FLOOD - anti flooding code for drones
 * This code adds server side ignore for a client who gets
 * messaged more than drone_count times within drone_time seconds
 * unfortunately, its a great DOS, but at least the client won't flood off.
 * I have no idea what to use for values here, trying 8 privmsgs
 * within 1 seconds. (I'm told it is usually that fast)
 * I'll do better next time, this is a Q&D -Dianora
 */
#define ANTI_DRONE_FLOOD
#define DEFAULT_DRONE_TIME 1
#define DEFAULT_DRONE_COUNT 5

#ifdef ANTI_DRONE_FLOOD
# define DRONETIME DEFAULT_DRONE_TIME
# define DRONECOUNT DEFAULT_DRONE_COUNT
#endif
/*
 * OLD_Y_LIMIT
 *
 * #define this if you prefer the old behaviour of I lines the default
 * behaviour is to limit the total number of clients using the max
 * client limit in the corresponding Y line (class) The old behaviour
 * was to limit the client count per I line without regard to the total
 * class limit. Each have advantages and disadvantages. In an open I
 * line server, the default behaviour i.e. #undef OLD_Y_LIMIT makes
 * more sense, because you can limit the total number of clients in a
 * class. In a closed I line server The old behaviour can make more
 * sense.
 *
 * -Dianora
 */
#undef OLD_Y_LIMIT

/*
 * If the OS has SOMAXCONN use that value, otherwise Use the value in
 * HYBRID_SOMAXCONN for the listen(); backlog try 5 or 25. 5 for AIX
 * and SUNOS, 25 should work better for other OS's
 */
#define HYBRID_SOMAXCONN 25

/*
 * Throttling support:
 * THROTTLE_ENABLE    - enable throttling code, if undefined, the functions will be empty. runtime settable.
 * THROTTLE_TRIGCOUNT - number of connections to triggle throttle action.
 * THROTTLE_TRIGTIME  - number of seconds in which THROTTLE_TRIGCOUNT must happen.
 * THROTTLE_RECORDTIME- length to keep records for each ip (since last connect from this ip).
 * THROTTLE_HASHSIZE  - size of the throttle hashtable, also tuneable.
 *
 * Recommended values: 3, 15, 1800.  3+ connections in 15 or less seconds will
 * result in a connection throttle z:line.  These are also
 * z: line time grows, pseudo-exponentially
 *  first zline : 2 minutes
 *  second zline: 5 minutes
 *  third zline : 15 minutes
 *  fourth zline: 30 minutes
 *  anything more is an hour
 * tuneable at runtime.  -wd */
#define THROTTLE_ENABLE
#define THROTTLE_TRIGCOUNT 3
#define THROTTLE_TRIGTIME 15
#define THROTTLE_RECORDTIME 1800
#define THROTTLE_HASHSIZE 25147

/*
 * ----------------- archaic and/or broken secion --------------------
 */

/*
 * USE_REJECT_HOLD clients that reconnect but are k-lined will have
 * their connections "held" for REJECT_HOLD_TIME seconds, they cannot
 * PRIVMSG. The idea is to keep a reconnecting client from forcing the
 * ircd to re-scan dich_conf. There is one possible serious hitch with
 * this... If it is a mass cloner, your attacker can "REJECT_HOLD" a
 * number of local fd's on your server. Against normal bots this code
 * will be a "win", against mass cloners, it could lose.
 *
 * this is still not working yet
 */
#undef USE_REJECT_HOLD
#define REJECT_HOLD_TIME 60

#undef DNS_DEBUG

/*
 * DEBUGMODE is used mostly for internal development, it is likely to
 * make your client server very sluggish. You usually shouldn't need
 * this. -Dianora
 */
#undef DEBUGMODE		/* define DEBUGMODE to enable */
#undef DUMP_DEBUG

/*
 * If your host supports varargs and has vsprintf(), vprintf() and
 * vscanf() C calls in its library, then you can define USE_VARARGS to
 * use varargs instead of imitation variable arg passing. #undef
 * USE_STDARGS
 *
 * NOTE: with current server code, varargs doesn't survive because it
 * can't be used in a chain of 3 or more funtions which all have a
 * variable number of params.  If anyone has a solution to this, please
 * notify the maintainer.
 *
 * don't use this right now, in a future release it will (should) work, but
 * now it doesn't */
#undef USE_STDARGS

/* ------------------------- END CONFIGURATION SECTION -------------------- */
/* ------------------ DO NOT TOUCH ANYTHING BELOW THIS LINE!!! ------------ */

#undef SYSLOG_BLOCK_ALLOCATOR  /* debug block allocator */
#undef USE_SYSLOG

#ifdef APOLLO
#define RESTARTING_SYSTEMCALLS
#endif /*
        * read/write are restarted after signals
        * defining this 1, gets siginterrupt call
        * compiled, which attempts to remove this
        * behaviour (apollo sr10.1/bsd4.3 needs this)
        */

#define HELPFILE HPATH
#define MOTD MPATH
#define OMOTD OMPATH
#define SHORTMOTD SMPATH
#define MYNAME SPATH
#define CONFIGFILE CPATH
#ifdef  KPATH
#define KLINEFILE KPATH
#endif
#ifdef VPATH
#define VLINEFILE VPATH
#endif

#define IRCD_PIDFILE PPATH
#define RULES RPATH

#define MAX_CLIENTS INIT_MAXCLIENTS

#if defined(CLIENT_FLOOD) && ((CLIENT_FLOOD > 8000) || (CLIENT_FLOOD < 512))
#error CLIENT_FLOOD needs redefining.
#endif

#if !defined(CLIENT_FLOOD)
#error CLIENT_FLOOD undefined.
#endif

#if (defined(OPTIONS_H) && !defined(OPTIONS_H_14))
#error Using outdated include/options.h - run config
#endif

#if defined(DEBUGMODE) || defined(DNS_DEBUG)
extern void debug(int level, char *pattern, ...);
#define Debug(x) debug x
#define LOGFILE SYSTEMPATH
#else
#define Debug(x) ;
#define LOGFILE "/dev/null"
#endif

#define CONFIG_H_LEVEL_12
#endif	/* __config_include__ */
