/*
 * VIAGRA Internet Relay Chat daemon, include/struct.h
 * Copyright (C) 1990 Jarkko Oikarinen and
 *                    University of Oulu, Computing Center
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef	__struct_include__
#define __struct_include__

#include "config.h"
#if !defined(CONFIG_H_LEVEL_12)
#error Incorrect config.h for this revision of ircd.
#endif
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#if defined( HAVE_STDDEF_H )
#include <stddef.h>
#endif
#ifdef ORATIMING
#include <sys/time.h>
#endif
#ifdef USE_SYSLOG
#include <syslog.h>
#if defined( HAVE_SYS_SYSLOG_H )
#include <sys/syslog.h>
#endif
#endif

#define REPORT_DO_DNS_ ":%s NOTICE AUTH :*** Looking up your hostname...\n"
#define REPORT_FIN_DNS_ ":%s NOTICE AUTH :*** Found your hostname\n"
#define REPORT_FIN_DNSC_ ":%s NOTICE AUTH :*** Found your hostname (cached)\n"
#define REPORT_FAIL_DNS_ ":%s NOTICE AUTH :*** Couldn't resolve your hostname; using your IP address instead\n"
#define REPORT_DO_ID_ ":%s NOTICE AUTH :*** Checking Ident...\n"
#define REPORT_FIN_ID_ ":%s NOTICE AUTH :*** Received identd response\n"
#define REPORT_FAIL_ID_ ":%s NOTICE AUTH :*** No ident response; username prefixed with ~\n"

extern char REPORT_DO_DNS[256], REPORT_FIN_DNS[256], REPORT_FIN_DNSC[256], REPORT_FAIL_DNS[256], REPORT_DO_ID[256], REPORT_FIN_ID[256], REPORT_FAIL_ID[256];

#include "hash.h"
typedef struct ConfItem aConfItem;
typedef struct Client aClient;
typedef struct Channel aChannel;
typedef struct User anUser;
typedef struct Server aServer;
typedef struct SLink Link;
typedef struct SLinkD DLink;
typedef struct ChanLink chanMember;
typedef struct SMode Mode;
typedef struct Watch aWatch;
typedef struct Ban aBan;
typedef struct Exempt anExempt;
typedef struct ListOptions LOpts;
typedef long ts_val;
typedef struct RulesItem aRules;
typedef struct MotdItem aMotd;
typedef struct OMotdItem aOMotd;
typedef struct t_crline aCRline;
typedef struct t_bline aBline;
typedef	struct Gline aGline;
typedef struct Sline aSline;

#include "class.h"
#include "dbuf.h"		/* THIS REALLY SHOULDN'T BE HERE!!! --msa */

#define	HOSTLEN 84		/* Length of hostname. Updated to comply with RFC1123 */
#define HOSTIPLEN 15		/* Length of dotted quad form of IP */
#define	NICKLEN 30
#define SM_MAXMODES 6
#define MAX_DATE_STRING 32	/* maximum string length for a date string */
#define	USERLEN 10
#define	REALLEN 50
#define	TOPICLEN 307
#define	KILLLEN	400
#define	CHANNELLEN 32
#define	PASSWDLEN 63
#define	KEYLEN 23
#define	BUFSIZE 512		/* WARNING: *DONT* CHANGE THIS!!!! */
#define	MAXBANS 50
#define MAXEXEMPTS 50
#define MOTDLINELEN 90
#define MAXSILES 5
#define MAXSILELENGTH 128
#define MAXDCCALLOW 5
#define DCC_LINK_ME 0x01	/* This is my dcc allow */
#define DCC_LINK_REMOTE 0x02    /* I need to remove these dcc allows from these clients when I die */

/*'offsetof' is defined in ANSI-C. The following definition is not
 * absolutely portable (I have been told), but so far it has worked
 * on all machines I have needed it. The type should be size_t but...
 * --msa */
#ifndef offsetof
#define	offsetof(t,m) (int)((&((t *)0L)->m))
#endif

#define	elementsof(x) (sizeof(x)/sizeof(x[0]))

/* flags for bootup options (command line flags) */
#define	BOOT_CONSOLE     1
#define	BOOT_QUICK	 2
#define	BOOT_DEBUG	 4
#define	BOOT_INETD	 8
#define	BOOT_TTY	 16
#define	BOOT_OPER	 32
#define	BOOT_AUTODIE     64
#define BOOT_STDERR	 128
#define	STAT_LOG	 -6	/* logfile for -x */
#define	STAT_MASTER	 -5	/* Local ircd master before identification */
#define	STAT_CONNECTING	 -4
#define	STAT_HANDSHAKE	 -3
#define	STAT_ME		 -2
#define	STAT_UNKNOWN	 -1
/* the line of truth lies here (truth == registeredness) */
#define	STAT_SERVER 0
#define	STAT_CLIENT 1

/* status macros. */
#define	IsRegisteredUser(x)	((x)->status == STAT_CLIENT)
#define	IsRegistered(x)		((x)->status >= STAT_SERVER)
#define	IsConnecting(x)		((x)->status == STAT_CONNECTING)
#define	IsHandshake(x)		((x)->status == STAT_HANDSHAKE)
#define	IsMe(x)			((x)->status == STAT_ME)
#define	IsUnknown(x)		((x)->status == STAT_UNKNOWN || (x)->status == STAT_MASTER)
#define	IsServer(x)		((x)->status == STAT_SERVER)
#define	IsClient(x)		((x)->status == STAT_CLIENT)
#define	IsLog(x)		((x)->status == STAT_LOG)

#define	SetMaster(x)		((x)->status = STAT_MASTER)
#define	SetConnecting(x)	((x)->status = STAT_CONNECTING)
#define	SetHandshake(x)		((x)->status = STAT_HANDSHAKE)
#define	SetMe(x)		((x)->status = STAT_ME)
#define	SetUnknown(x)		((x)->status = STAT_UNKNOWN)
#define	SetServer(x)		((x)->status = STAT_SERVER)
#define	SetClient(x)		((x)->status = STAT_CLIENT)
#define	SetLog(x)		((x)->status = STAT_LOG)

#define	FLAGS_PINGSENT     0x000001   /* Unreplied ping sent */
#define	FLAGS_DEADSOCKET   0x000002   /* Local socket is dead--Exiting soon */
#define	FLAGS_KILLED       0x000004   /* Prevents "QUIT" from being sent for this */
#define	FLAGS_BLOCKED      0x000008   /* socket is in a blocked condition */
#define FLAGS_REJECT_HOLD  0x000010   /* client has been klined */
#define	FLAGS_CLOSING      0x000020   /* set when closing to suppress errors */
#define	FLAGS_LISTEN       0x000040   /* used to mark clients which we listen() on */
#define	FLAGS_CHKACCESS    0x000080   /* ok to check clients access if set */
#define	FLAGS_DOINGDNS	   0x000100   /* client is waiting for a DNS response */
#define	FLAGS_AUTH	   0x000200   /* client is waiting on rfc931 response */
#define	FLAGS_WRAUTH	   0x000400   /* set if we havent writen to ident server */
#define	FLAGS_LOCAL	   0x000800   /* set for local clients */
#define	FLAGS_GOTID	   0x001000   /* successful ident lookup achieved */
#define	FLAGS_DOID	   0x002000   /* I-lines say must use ident return */
#define	FLAGS_NONL	   0x004000   /* No \n in buffer */
#define FLAGS_NORMALEX     0x008000   /* Client exited normally */
#define FLAGS_SENDQEX      0x010000   /* Sendq exceeded */
#define FLAGS_IPHASH       0x020000   /* iphashed this client */
#define FLAGS_ULINE 	   0x040000   /* client is U-lined */
#define FLAGS_USERBURST	   0x080000   /* server in nick/channel netburst */
#define FLAGS_TOPICBURST   0x100000   /* server in topic netburst */
#define FLAGS_BURST	   (FLAGS_USERBURST | FLAGS_TOPICBURST)
#define FLAGS_SOBSENT      0x200000   /* we've sent an SOB, just have to send an EOB */
#define FLAGS_EOBRECV      0x400000   /* we're waiting on an EOB */
#define FLAGS_BAD_DNS	   0x800000   /* spoofer-guy */
#define FLAGS_SERV_NEGO	   0x1000000  /* This is a server that has passed connection tests, but is a stat < 0 for handshake purposes */
#define FLAGS_RC4IN        0x2000000  /* This link is rc4 encrypted. */
#define FLAGS_RC4OUT       0x4000000  /* This link is rc4 encrypted. */
#define FLAGS_ZIPPED_IN	   0x8000000  /* This link is gzipped. */
#define FLAGS_ZIPPED_OUT   0x10000000 /* This link is gzipped. */
#define FLAGS_MAP          0x20000000
#define FLAGS2_TCTRL	   0x00000001

/* Capabilities of the ircd or clients */
#define CAPAB_TS3     0x0000001	/* Supports the TS3 Protocol */
#define CAPAB_TS5     0x0000002 /* Supports the TS5 Protocol */
#define CAPAB_NOQUIT  0x0000004 /* Supports NOQUIT */
#define CAPAB_NSJOIN  0x0000008 /* server supports new smart sjoin */
#define CAPAB_BURST   0x0000010 /* server supports BURST command */
#define CAPAB_UNCONN  0x0000020 /* server supports UNCONNECT */
#define CAPAB_DKEY    0x0000040 /* server supports dh-key exchange using "DKEY" */
#define CAPAB_ZIP     0x0000080 /* server supports gz'd links */
#define CAPAB_DOZIP   0x0000100 /* output to this link shall be gzipped */
#define CAPAB_DODKEY  0x0000200 /* do I do dkey with this link? */
#define CAPAB_NICKIP  0x0000400 /* IP in the NICK line? */
#define CAPAB_TSMODE  0x0000400 /* MODE's parv[2] is chptr->channelts for channel mode */

#define SetTS5(x) ((x)->capabilities |= CAPAB_TS5)
#define IsTS5(x)   ((x)->capabilities & CAPAB_TS5)
#define SetTS3(x) ((x)->capabilities |= CAPAB_TS3)
#define IsTS3(x)   ((x)->capabilities & CAPAB_TS3)
#define SetNoQuit(x) 	((x)->capabilities |= CAPAB_NOQUIT)
#define IsNoQuit(x) 	 ((x)->capabilities & CAPAB_NOQUIT)
#define SetSSJoin(x)	((x)->capabilities |= CAPAB_NSJOIN)
#define IsSSJoin(x)	 ((x)->capabilities & CAPAB_NSJOIN)
#define SetBurst(x)	((x)->capabilities |= CAPAB_BURST)
#define IsBurst(x)	 ((x)->capabilities & CAPAB_BURST)
#define SetUnconnect(x)	((x)->capabilities |= CAPAB_UNCONN)
#define IsUnconnect(x)	 ((x)->capabilities & CAPAB_UNCONN)
#define SetDKEY(x)	((x)->capabilities |= CAPAB_DKEY)
#define CanDoDKEY(x)     ((x)->capabilities & CAPAB_DKEY)
/* N: line, flag E */
#define WantDKEY(x)    ((x)->capabilities & CAPAB_DODKEY) 

#define SetZipCapable(x) ((x)->capabilities |= CAPAB_ZIP)
#define IsZipCapable(x)	  ((x)->capabilities & CAPAB_ZIP)
/* this is set in N: line, flag Z */
#define DoZipThis(x)    ((x)->capabilities & CAPAB_DOZIP) 
#define SetNICKIP(x)  ((x)->capabilities |= CAPAB_NICKIP)
#define IsNICKIP(x)    ((x)->capabilities & CAPAB_NICKIP)
#define SetTSMODE(x)	((x)->capabilities |= CAPAB_TSMODE)
#define IsTSMODE(x)	((x)->capabilities & CAPAB_TSMODE)

/* flag macros. */
#define IsULine(x) ((x)->flags & FLAGS_ULINE)

/* futured usage *
#define UMODE_a  1
#define UMODE_b  2
#define UMODE_c  3
#define UMODE_d  4
#define UMODE_e  5
#define UMODE_f  6
#define UMODE_g  7
#define UMODE_h  8
#define UMODE_i  9
#define UMODE_j 10
#define UMODE_k 11
#define UMODE_l 12
#define UMODE_m 13
#define UMODE_n 14
#define UMODE_o 15
#define UMODE_p 16
#define UMODE_q 17
#define UMODE_r 18
#define UMODE_s 19
#define UMODE_t 20
#define UMODE_u 21
#define UMODE_v 22
#define UMODE_w 23
#define UMODE_x 24
#define UMODE_y 25
#define UMODE_z 26
#define UMODE_A 27
#define UMODE_B 28
#define UMODE_C 29
#define UMODE_D 30
#define UMODE_E 31
#define UMODE_F 32
#define UMODE_G 33
#define UMODE_H 34
#define UMODE_I 35
#define UMODE_J 36
#define UMODE_K 37
#define UMODE_L 38
#define UMODE_M 39
#define UMODE_N 40
#define UMODE_O 41
#define UMODE_P 42
#define UMODE_Q 43
#define UMODE_R 44
#define UMODE_S 45
#define UMODE_T 46
#define UMODE_U 47
#define UMODE_V 48
#define UMODE_W 49
#define UMODE_X 50
#define UMODE_Y 51
#define UMODE_Z 52
#define UMODE_INVALID 0

#define HasUmodea(x) (HasUmode(x,UMODE_a))
#define HasUmodeb(x) (HasUmode(x,UMODE_b))
#define HasUmodec(x) (HasUmode(x,UMODE_c))
#define HasUmoded(x) (HasUmode(x,UMODE_d))
#define HasUmodee(x) (HasUmode(x,UMODE_e))
#define HasUmodef(x) (HasUmode(x,UMODE_f))
#define HasUmodeg(x) (HasUmode(x,UMODE_g))
#define HasUmodeh(x) (HasUmode(x,UMODE_h))
#define HasUmodei(x) (HasUmode(x,UMODE_i))
#define HasUmodej(x) (HasUmode(x,UMODE_j))
#define HasUmodek(x) (HasUmode(x,UMODE_k))
#define HasUmodel(x) (HasUmode(x,UMODE_l))
#define HasUmodem(x) (HasUmode(x,UMODE_m))
#define HasUmoden(x) (HasUmode(x,UMODE_n))
#define HasUmodeo(x) (HasUmode(x,UMODE_o))
#define HasUmodep(x) (HasUmode(x,UMODE_p))
#define HasUmodeq(x) (HasUmode(x,UMODE_q))
#define HasUmoder(x) (HasUmode(x,UMODE_r))
#define HasUmodes(x) (HasUmode(x,UMODE_s))
#define HasUmodet(x) (HasUmode(x,UMODE_t))
#define HasUmodeu(x) (HasUmode(x,UMODE_u))
#define HasUmodev(x) (HasUmode(x,UMODE_v))
#define HasUmodew(x) (HasUmode(x,UMODE_w))
#define HasUmodex(x) (HasUmode(x,UMODE_x))
#define HasUmodey(x) (HasUmode(x,UMODE_y))
#define HasUmodez(x) (HasUmode(x,UMODE_z))
#define HasUmodeA(x) (HasUmode(x,UMODE_A))
#define HasUmodeB(x) (HasUmode(x,UMODE_B))
#define HasUmodeC(x) (HasUmode(x,UMODE_C))
#define HasUmodeD(x) (HasUmode(x,UMODE_D))
#define HasUmodeE(x) (HasUmode(x,UMODE_E))
#define HasUmodeF(x) (HasUmode(x,UMODE_F))
#define HasUmodeG(x) (HasUmode(x,UMODE_G))
#define HasUmodeH(x) (HasUmode(x,UMODE_H))
#define HasUmodeI(x) (HasUmode(x,UMODE_I))
#define HasUmodeJ(x) (HasUmode(x,UMODE_J))
#define HasUmodeK(x) (HasUmode(x,UMODE_K))
#define HasUmodeL(x) (HasUmode(x,UMODE_L))
#define HasUmodeM(x) (HasUmode(x,UMODE_M))
#define HasUmodeN(x) (HasUmode(x,UMODE_N))
#define HasUmodeO(x) (HasUmode(x,UMODE_O))
#define HasUmodeP(x) (HasUmode(x,UMODE_P))
#define HasUmodeQ(x) (HasUmode(x,UMODE_Q))
#define HasUmodeR(x) (HasUmode(x,UMODE_R))
#define HasUmodeS(x) (HasUmode(x,UMODE_S))
#define HasUmodeT(x) (HasUmode(x,UMODE_T))
#define HasUmodeU(x) (HasUmode(x,UMODE_U))
#define HasUmodeV(x) (HasUmode(x,UMODE_V))
#define HasUmodeW(x) (HasUmode(x,UMODE_W))
#define HasUmodeX(x) (HasUmode(x,UMODE_X))
#define HasUmodeY(x) (HasUmode(x,UMODE_Y))
#define HasUmodeZ(x) (HasUmode(x,UMODE_Z))
***					*/

/* sendto_umode() levels */
#define MODE_a 1 /* */
#define MODE_b 2 /* */
#define MODE_c 3 /* */
#define MODE_d 4 /* */
#define MODE_e 5 /* */
#define MODE_f 6 /* */
#define MODE_k 7 /* */
#define MODE_n 8 /* */
#define MODE_s 9 /* */

/*		   User Modes */
#define UMODE_o 0x00000001 /* */
#define UMODE_O 0x00000002 /* */
#define UMODE_i 0x00000004 /* */
#define UMODE_w 0x00000008 /* */
#define UMODE_s 0x00000010 /* */
#define UMODE_c 0x00000020 /* */
#define UMODE_r 0x00000040 /* */
#define UMODE_k 0x00000080 /* */
#define UMODE_f 0x00000100 /* */
#define UMODE_W 0x00000200 /* */
#define UMODE_d 0x00000400 /* */
#define UMODE_I 0x00000800 /* */
#define UMODE_g 0x00001000 /* */
#define UMODE_p 0x00002000 /* */
#define UMODE_a 0x00004000 /* */
#define UMODE_A 0x00008000 /* */
#define UMODE_n 0x00010000 /* */
#define UMODE_h 0x00020000 /* */
#define UMODE_m 0x00040000 /* */
#define UMODE_R 0x00080000 /* */
#define UMODE_b 0x00100000 /* */
#define UMODE_e 0x00200000 /* */
#define UMODE_q 0x00400000 /* */
#define UMODE_D 0x00800000 /* */
#define UMODE_x 0x01000000 /* */
#define UMODE_S 0x02000000 /* */
#define UMODE_N 0x04000000 /* */
#define UMODE_H 0x08000000 /* */
#define UMODE_T 0x10000000 /* */
#define UMODE_X 0x20000000 /* */
#define UMODE_Q 0x40000000 /* */
#define UMODE_C 0x80000000 /* */

#define	SEND_UMODES (UMODE_a|UMODE_b|UMODE_x|UMODE_o|UMODE_w|UMODE_r|UMODE_i|UMODE_A|UMODE_W|UMODE_e|UMODE_d|\
                     UMODE_h|UMODE_R|UMODE_S|UMODE_q|UMODE_C|UMODE_Q|UMODE_X|UMODE_H|UMODE_N|UMODE_T|UMODE_I|UMODE_D)
#define ALL_UMODES (SEND_UMODES|UMODE_s|UMODE_b|UMODE_c|UMODE_r|UMODE_k|UMODE_f|UMODE_W|UMODE_d|UMODE_g|UMODE_n|UMODE_h|UMODE_m|\
	            UMODE_O|UMODE_R|UMODE_e|UMODE_W|UMODE_N|UMODE_T|UMODE_X|UMODE_H|UMODE_Q|UMODE_C|UMODE_I|UMODE_q|UMODE_D)
#define	FLAGS_ID (FLAGS_DOID|FLAGS_GOTID)

#define HasUmodea(x) ((x)->umode & UMODE_a) /* */
#define HasUmodeb(x) ((x)->umode & UMODE_b) /* */
#define HasUmodec(x) ((x)->umode & UMODE_c) /* */
#define HasUmoded(x) ((x)->umode & UMODE_d) /* */
#define HasUmodee(x) ((x)->umode & UMODE_e) /* */
#define HasUmodef(x) ((x)->umode & UMODE_f) /* */
#define HasUmodeg(x) ((x)->umode & UMODE_g) /* */
#define HasUmodeh(x) ((x)->umode & UMODE_h) /* */
#define HasUmodei(x) ((x)->umode & UMODE_i) /* */
#define HasUmodek(x) ((x)->umode & UMODE_k) /* */
#define HasUmodeo(x) ((x)->umode & UMODE_o) /* */
#define HasUmodem(x) ((x)->umode & UMODE_m) /* */
#define HasUmoden(x) ((x)->umode & UMODE_n) /* */
#define HasUmodeq(x) ((x)->umode & UMODE_q) /* */
#define HasUmoder(x) ((x)->umode & UMODE_r) /* */
#define HasUmodes(x) ((x)->umode & UMODE_s) /* */
#define HasUmodew(x) ((x)->umode & UMODE_w) /* */
#define HasUmodex(x) ((x)->umode & UMODE_x) /* */
#define HasUmodeA(x) ((x)->umode & UMODE_A) /* */
#define HasUmodeC(x) ((x)->umode & UMODE_C) /* */
#define HasUmodeD(x) ((x)->umode & UMODE_D) /* */
#define HasUmodeH(x) ((x)->umode & UMODE_H) /* */
#define HasUmodeI(x) ((x)->umode & UMODE_I) /* */
#define HasUmodeN(x) ((x)->umode & UMODE_N) /* */
#define HasUmodeO(x) ((x)->umode & UMODE_O) /* */
#define HasUmodeQ(x) ((x)->umode & UMODE_Q) /* */
#define HasUmodeR(x) ((x)->umode & UMODE_R) /* */
#define HasUmodeS(x) ((x)->umode & UMODE_S) /* */
#define HasUmodeT(x) ((x)->umode & UMODE_T) /* */
#define HasUmodeW(x) ((x)->umode & UMODE_W) /* */
#define HasUmodeX(x) ((x)->umode & UMODE_X) /* */

#define SetUmodea(x) ((x)->umode |= UMODE_a) /* */
#define SetUmodeb(x) ((x)->umode |= UMODE_b) /* */
#define SetUmodec(x) ((x)->umode |= UMODE_c) /* */
#define SetUmodee(x) ((x)->umode |= UMODE_e) /* */
#define SetUmodef(x) ((x)->umode |= UMODE_f) /* */
#define SetUmodeg(x) ((x)->umode |= UMODE_g) /* */
#define SetUmodeh(x) ((x)->umode |= UMODE_h) /* */
#define SetUmodei(x) ((x)->umode |= UMODE_i) /* */
#define SetUmodek(x) ((x)->umode |= UMODE_k) /* */
#define SetUmoden(x) ((x)->umode |= UMODE_n) /* */
#define SetUmodeo(x) ((x)->umode |= UMODE_o) /* */
#define SetUmodes(x) ((x)->umode |= UMODE_s) /* */
#define SetUmodew(x) ((x)->umode |= UMODE_w) /* */
#define SetUmodex(x) ((x)->umode |= UMODE_x) /* */
#define SetUmodeA(x) ((x)->umode |= UMODE_A) /* */
#define SetUmodeC(x) ((x)->umode |= UMODE_C) /* */
#define SetUmodeD(x) ((x)->umode |= UMODE_D) /* */
#define SetUmodeH(x) ((x)->umode |= UMODE_H) /* */
#define SetUmodeI(x) ((x)->umode |= UMODE_I) /* */
#define SetUmodeN(x) ((x)->umode |= UMODE_N) /* */
#define SetUmodeO(x) ((x)->umode |= UMODE_O) /* */
#define SetUmodeQ(x) ((x)->umode |= UMODE_Q) /* */
#define SetUmodeR(x) ((x)->umode |= UMODE_R) /* */
#define SetUmodeT(x) ((x)->umode |= UMODE_T) /* */
#define SetUmodeW(x) ((x)->umode |= UMODE_W) /* */
#define SetUmodeX(x) ((x)->umode |= UMODE_X) /* */

#define ClearUmodea(x) ((x)->umode &= ~UMODE_a) /* */
#define ClearUmodeb(x) ((x)->umode &= ~UMODE_b) /* */
#define ClearUmodec(x) ((x)->umode &= ~UMODE_c) /* */ 
#define ClearUmoded(x) ((x)->umode &= ~UMODE_d) /* */
#define ClearUmodee(x) ((x)->umode &= ~UMODE_e) /* */
#define ClearUmodef(x) ((x)->umode &= ~UMODE_f) /* */
#define ClearUmodeh(x) ((x)->umode &= ~UMODE_h) /* */
#define ClearUmodei(x) ((x)->umode &= ~UMODE_i) /* */
#define ClearUmodek(x) ((x)->umode &= ~UMODE_k) /* */
#define ClearUmodem(x) ((x)->umode &= ~UMODE_m) /* */
#define ClearUmoden(x) ((x)->umode &= ~UMODE_n) /* */
#define ClearUmodeo(x) ((x)->umode &= ~UMODE_o) /* */
#define ClearUmodeq(x) ((x)->umode &= ~UMODE_q) /* */
#define ClearUmodes(x) ((x)->umode &= ~UMODE_s) /* */
#define ClearUmodex(x) ((x)->umode &= ~UMODE_x) /* */
#define ClearUmodeA(x) ((x)->umode &= ~UMODE_A) /* */
#define ClearUmodeC(x) ((x)->umode &= ~UMODE_C) /* */
#define ClearUmodeD(x) ((x)->umode &= ~UMODE_D) /* */
#define ClearUmodeH(x) ((x)->umode &= ~UMODE_H) /* */
#define ClearUmodeI(x) ((x)->umode &= ~UMODE_I) /* */
#define ClearUmodeN(x) ((x)->umode &= ~UMODE_N) /* */
#define ClearUmodeO(x) ((x)->umode &= ~UMODE_O) /* */
#define ClearUmodeQ(x) ((x)->umode &= ~UMODE_Q) /* */
#define ClearUmodeS(x) ((x)->umode &= ~UMODE_S) /* */
#define ClearUmodeT(x) ((x)->umode &= ~UMODE_T) /* */
#define ClearUmodeW(x) ((x)->umode &= ~UMODE_W) /* */
#define ClearUmodeX(x) ((x)->umode &= ~UMODE_X) /* */

#define OFlagA(x) ((x)->oflag & OFLAG_A)
#define OFlagD(x) ((x)->oflag & OFLAG_D)
#define OFlaga(x) ((x)->oflag & OFLAG_a)
#define OFlagg(x) ((x)->oflag & OFLAG_g)
#define OFlagR(x) ((x)->oflag & OFLAG_R)
#define OFlagT(x) ((x)->oflag & OFLAG_T)
#define OFlagN(x) ((x)->oflag & OFLAG_N)
#define OFlagX(x) ((x)->oflag & OFLAG_X)
#define OFlagC(x) ((x)->oflag & OFLAG_C)
#define OFlagc(x) ((x)->oflag & OFLAG_c)
#define OFlagQ(x) ((x)->oflag & OFLAG_Q)
#define OFlagH(x) ((x)->oflag & OFLAG_H)
#define OFlagI(x) ((x)->oflag & OFLAG_I)
#define OFlagW(x) ((x)->oflag & OFLAG_W)
#define OFlage(x) ((x)->oflag & OFLAG_e)
#define OFlagr(x) ((x)->oflag & OFLAG_r)
#define OFlagu(x) ((x)->oflag & OFLAG_u)
#define OFlagz(x) ((x)->oflag & OFLAG_z)

#define	IsOper(x) ((x)->umode & UMODE_o)
#define	IsAnOper(x)		((x)->umode & (UMODE_o|UMODE_O))
#define IsARegNick(x)           ((x)->umode & (UMODE_r))
#define IsRegNick(x)            ((x)->umode & UMODE_r)
#define IsSAdmin(x)             ((x)->umode & UMODE_a)
#define IsAdmin(x)              ((x)->umode & UMODE_A)

#define	IsPerson(x)		((x)->user && IsClient(x))
#define	IsPrivileged(x)		(IsAnOper(x) || IsServer(x))

#define	IsListening(x)		((x)->flags & FLAGS_LISTEN)
#define	DoAccess(x)		((x)->flags & FLAGS_CHKACCESS)
#define	IsLocal(x)		((x)->flags & FLAGS_LOCAL)
#define	IsDead(x)		((x)->flags & FLAGS_DEADSOCKET)

#define	SetDNS(x)		((x)->flags |= FLAGS_DOINGDNS)
#define	DoingDNS(x)		((x)->flags & FLAGS_DOINGDNS)
#define	SetAccess(x)		((x)->flags |= FLAGS_CHKACCESS)
#define	DoingAuth(x)		((x)->flags & FLAGS_AUTH)
#define	NoNewLine(x)		((x)->flags & FLAGS_NONL)

#define GotTctrl(x)  ((x)->flags2 & FLAGS2_TCTRL)
#define SetTctrl(x) ((x)->flags2 |= FLAGS2_TCTRL)

#define SetNegoServer(x)	((x)->flags |= FLAGS_SERV_NEGO)
#define IsNegoServer(x)		((x)->flags & FLAGS_SERV_NEGO)
#define ClearNegoServer(x)	((x)->flags &= ~FLAGS_SERV_NEGO)
#define IsRC4OUT(x)		((x)->flags & FLAGS_RC4OUT)
#define SetRC4OUT(x)		((x)->flags |= FLAGS_RC4OUT)
#define IsRC4IN(x)		((x)->flags & FLAGS_RC4IN)
#define SetRC4IN(x)		((x)->flags |= FLAGS_RC4IN)
#define RC4EncLink(x)		(((x)->flags & (FLAGS_RC4IN|FLAGS_RC4OUT)) == (FLAGS_RC4IN|FLAGS_RC4OUT))

#define ZipIn(x)		((x)->flags & FLAGS_ZIPPED_IN)
#define SetZipIn(x)		((x)->flags |= FLAGS_ZIPPED_IN)
#define ZipOut(x)		((x)->flags & FLAGS_ZIPPED_OUT)
#define SetZipOut(x)		((x)->flags |= FLAGS_ZIPPED_OUT)

#define	ClearDNS(x)		((x)->flags &= ~FLAGS_DOINGDNS)
#define	ClearAuth(x)		((x)->flags &= ~FLAGS_AUTH)
#define	ClearAccess(x)		((x)->flags &= ~FLAGS_CHKACCESS)

/* defined operator access levels */ 
#define OFLAG_r 0x00000001 /* */
#define OFLAG_D 0x00000002 /* */
#define OFLAG_R 0x00000004 /* */
#define OFLAG_v 0x00000008 /* add glines */
#define OFLAG_h 0x00000010 /* */
#define OFLAG_g 0x00000020 /* */
#define OFLAG_w 0x00000040 /* */
#define OFLAG_l 0x00000080 /* */
#define OFLAG_c 0x00000100 /* */
#define OFLAG_L 0x00000200 /* */
#define OFLAG_k 0x00000400 /* */
#define OFLAG_K 0x00000800 /* */
#define OFLAG_b 0x00001000 /* add klines */
#define OFLAG_B 0x00002000 /* unkline */
#define OFLAG_n 0x00004000 /* */
#define OFLAG_G 0x00008000 /* */
#define OFLAG_A 0x00010000 /* */
#define OFLAG_u 0x00020000 /* */
#define OFLAG_a 0x00080000 /* */
#define OFLAG_z 0x00100000 /* */
#define OFLAG_W 0x00200000 /* */
#define OFLAG_e 0x00400000 /* */
#define OFLAG_I 0x00800000 /* */
#define OFLAG_Y 0x01000000 /* */
#define OFLAG_N 0x02000000 /* */
#define OFLAG_T 0x04000000 /* */
#define OFLAG_X 0x08000000 /* */
#define OFLAG_Q 0x10000000 /* */
#define OFLAG_C 0x20000000 /* */
#define OFLAG_H 0x40000000 /* */
#define OFLAG_V 0x80000000 /* remove glines */

#define OFLAG_o (OFLAG_h|OFLAG_l|OFLAG_k|OFLAG_n|OFLAG_u)
#define OFLAG_O (OFLAG_o|OFLAG_K|OFLAG_G|OFLAG_w|OFLAG_g|OFLAG_b|OFLAG_B|OFLAG_r)
#define OFLAG_ISGLOBAL  (OFLAG_K|OFLAG_G)

#define OPCanWallOps(x)	        ((x)->oflag & OFLAG_w)
#define OPCanLocOps(x)	        ((x)->oflag & OFLAG_l)
#define OPCanLRoute(x)	        ((x)->oflag & OFLAG_L)
#define OPCanGRoute(x)	        ((x)->oflag & OFLAG_Y)
#define OPCanLKill(x)	        ((x)->oflag & OFLAG_k)
#define OPCanGKill(x)	        ((x)->oflag & OFLAG_K)
#define OPCanAddGline(x)        ((x)->oflag & OFLAG_v)
#define OPCanDelGline(x)        ((x)->oflag & OFLAG_V)
#define OPCanKline(x)	        ((x)->oflag & OFLAG_b)
#define OPCanUnKline(x)	        ((x)->oflag & OFLAG_B)
#define OPCanLNotice(x)	        ((x)->oflag & OFLAG_n)
#define OPCanGNotice(x)	        ((x)->oflag & OFLAG_G)

/* defined debugging levels */
#define	DEBUG_FATAL  0
#define	DEBUG_ERROR  1 /* report_error() and other errors that are found */
#define	DEBUG_NOTICE 3
#define	DEBUG_DNS    4 /* used by all DNS related routines - a *lot* */
#define	DEBUG_INFO   5 /* general usful info */
#define	DEBUG_NUM    6 /* numerics */
#define	DEBUG_SEND   7 /* everything that is sent out */
#define	DEBUG_DEBUG  8 /* anything to do with debugging, ie unimportant :) */
#define	DEBUG_MALLOC 9 /* malloc/free calls */
#define	DEBUG_LIST  10 /* debug list use */
/* defines for curses in client */
#define	DUMMY_TERM	0
#define	CURSES_TERM	1
#define	TERMCAP_TERM	2

struct Counter 
{
    int server;		/* servers */
    int myserver;	/* my servers */
    int myulined;	/* my ulined servers */
    int oper;		/* Opers */
    int chan;		/* Channels */
    int local;		/* Local Clients */
    int total;		/* total clients */
    int invisi;		/* invisible clients */
    int unknown;	/* unknown connections */
    int max_loc;	/* MAX local clients */
    int max_tot;	/* MAX global clients */
    unsigned long totalrestartcount;	/* Total clients since restart */
    ts_val start;	/* when we started collecting info */
    u_long today;	/* Client Connections today */
    ts_val day;		/* when today started */
    u_long weekly;	/* connections this week */
    ts_val week;	/* when this week started */
    u_long monthly;	/* connections this month */
    ts_val month;	/* when this month started */
    u_long yearly;	/* this is gonna be big */
    ts_val year;	/* when this year started (HEH!) */
};

struct MotdItem 
{
    char line[MOTDLINELEN];
    struct MotdItem *next;
};

struct OMotdItem
{
    char line[MOTDLINELEN];
    struct OMotdItem *next;
};

struct RulesItem
{
    char line[MOTDLINELEN];
    struct RulesItem *next;
};

/* lets speed this up... also removed away information. *tough* Dianora */
typedef struct Whowas 
{
    int hashv;
    char name[NICKLEN+1];
    char username[USERLEN+1];
    char hostname[HOSTLEN+1];
    char hostipname[HOSTLEN+1];
    char virthost[HOSTLEN+1];
    char *servername;
    char realname[REALLEN+1];
    time_t logoff;
    struct Client *online;	/* Pointer to new nickname for chasing or NULL */
    struct Whowas *next;	/* for hash table... */
    struct Whowas *prev;	/* for hash table... */
    struct Whowas *cnext;	/* for client struct linked list */
    struct Whowas *cprev;	/* for client struct linked list */
} aWhowas;

struct ConfItem 
{
    unsigned int status;	/* If CONF_ILLEGAL, delete when no clients */
    unsigned int flags;		/* i-lines and akills use this */
    int clients;		/* Number of *LOCAL* clients using this */
    struct in_addr ipnum;	/* ip number of host field */
    char *host;
    char *localhost;
    char *passwd;
    char *name;
    int port;
    time_t hold;		/* Hold action until this time (calendar time) */
    aClass *class;		/* Class of connection */
    struct ConfItem *next;
};

#define	CONF_ILLEGAL	        0x80000000
#define	CONF_MATCH	        0x40000000

#define	CONF_CLIENT	        0x0002
#define	CONF_CONNECT_SERVER	0x0004
#define	CONF_NOCONNECT_SERVER	0x0008
#define	CONF_LOCOP		0x0010
#define	CONF_OPERATOR		0x0020
#define	CONF_ME			0x0040
#define	CONF_KILL		0x0080
#define	CONF_ADMIN		0x0100
#define	CONF_CLASS		0x0200
#define	CONF_SERVICE		0x0400
#define CONF_VHOST		0x0800
#define	CONF_LISTEN_PORT	0x1000
#define	CONF_HUB		0x2000
#define CONF_ELINE		0x4000
#define CONF_FLINE		0x8000
#define	CONF_ZLINE		0x20000
#define CONF_QUARANTINED_NICK 	0x40000
#define CONF_ULINE 		0x80000
#define CONF_DRPASS		0x100000    /* die/restart pass, from df465 */
#define CONF_AKILL		0x200000
#define CONF_SQLINE     	0x400000
#define CONF_GCOS               0x1000000
#define CONF_SGLINE             0x2000000
#define CONF_SZLINE             0x4000000   /* Services placed zlines */
#define CONF_QUARANTINED_CHAN	0x8000000
#define CONF_QUARANTINE         (CONF_QUARANTINED_NICK|CONF_QUARANTINED_CHAN)
#define	CONF_OPS		(CONF_OPERATOR | CONF_LOCOP)
#define	CONF_SERVER_MASK	(CONF_CONNECT_SERVER | CONF_NOCONNECT_SERVER)
#define	CONF_CLIENT_MASK	(CONF_CLIENT | CONF_SERVICE | CONF_OPS | CONF_SERVER_MASK)
#define	IsIllegal(x)	        ((x)->status & CONF_ILLEGAL)

#define CONF_FLAGS_I_OPERPORT      0x0002 /* did the password field specify OPER? */
#define CONF_FLAGS_I_NAME_HAS_AT   0x0004 /* does NAME in I:HOST::NAME have an @? */
#define CONF_FLAGS_I_HOST_HAS_AT   0x0008 /* does HOST in I:HOST::NAME have an @? */

/* Client structures */
struct User
{
    Link *channel;		/* chain of channel pointer blocks */
    Link *invited;		/* chain of invite pointer blocks */
    char *away;			/* pointer to away message */
    time_t last;
    int joined;			/* number of channels joined */
    char username[USERLEN+1];
    char host[HOSTLEN+1];
    char virthost[HOSTLEN+1];
    char swhois[HOSTLEN+1];
    char *server;		/* pointer to scached server name */
#ifdef OS_SOLARIS
    uint_t servicestamp;	/* solaris is gay -epi */
#else
    u_int32_t servicestamp;	/* Services id - Raistlin */
#endif

    /* In a perfect world the 'server' name should not be needed, a
     * pointer to the client describing the server is enough. 
     * Unfortunately, in reality, server may not yet be in links while
     * USER is introduced... --msa */

    Link *silence;		/* chain of silenced users */
    LOpts *lopt;		/* Saved /list options */
    Link *dccallow;		/* chain of dcc send allowed users */
    char *real_oper_host;
    char *real_oper_username;
    char *real_oper_ip;
};

struct Server
{
    char *up;			/* Pointer to scache name */
    char bynick[NICKLEN+1];
    char byuser[USERLEN+1];
    char byhost[HOSTLEN+1];
    aConfItem *nline;		/* N-line pointer for this server */
    int dkey_flags;		/* dkey flags */
#ifdef HAVE_ENCRYPTION_ON
    void *sessioninfo_in;	/* pointer to opaque sessioninfo structure */
    void *sessioninfo_out;	/* pointer to opaque sessioninfo structure */
    void *rc4_in;		/* etc */
    void *rc4_out;		/* etc */
#endif
    void *zip_out;
    void *zip_in;
};

struct Client 
{
    struct Client *next, *prev, *hnext;
    anUser *user;		/* ...defined, if this is a User */
    aServer *serv;		/* ...defined, if this is a server */
    aWhowas *whowas;		/* Pointers to whowas structs */
    time_t lasttime;		/* ...should be only LOCAL clients? --msa */
    time_t firsttime;		/* time client was created */
    time_t since;		/* last time we parsed something */
    ts_val tsinfo;		/* TS on the nick, SVINFO on servers */
    long flags;			/* client flags */
    long flags2;		/* More client flags. */
    long umode;			/* We can illeviate overflow this way */
    aClient *from;		/* == self, if Local Client, *NEVER* NULL! */
    aClient *uplink;		/* this client's uplink to the network */
    int fd;			/* >= 0, for local clients */
    int hopcount;		/* number of servers to this 0 = local */
    short status;		/* Client type */
    char nicksent;
    char name[HOSTLEN+1];	/* Unique name of the client, nick or host */
    char info[REALLEN+1];	/* Free form additional client information */
#ifdef FLUD
    Link *fludees;
#endif
    struct in_addr ip;		/* keep real ip# too */
    char hostip[HOSTIPLEN+1];	/* Keep real ip as string too - Dianora */
    
    Link *watch;		/* user's watch list */
    int watches;		/* how many watches this user has set */
    
    /* The following fields are allocated only for local clients 
     * (directly connected to this server with a socket.  The first
     * of them MUST be the "count"--it is the field to which the
     * allocation is tied to! Never refer to  these fields, if (from != self). */
    
    int count;			/* Amount of data in buffer */
#ifdef FLUD
    time_t fludblock;
    struct fludbot *fluders;
#endif
    time_t last_join_time;	/* when this client last joined a channel */
    time_t last_leave_time;	/* when this client last left a channel */
    int join_leave_count;	/* count of JOIN/LEAVE in less than MIN_JOIN_LEAVE_TIME seconds */
    int oper_warn_count_down;	/* warn opers of this possible spambot every time this gets to 0 */
#ifdef ANTI_DRONE_FLOOD
    time_t first_received_message_time;
    int received_number_of_privmsgs;
    int drone_noticed;
#endif
    char buffer[BUFSIZE];	/* Incoming message buffer */
    short lastsq;		/* # of 2k blocks when sendqueued called last */
    struct DBuf sendQ;		/* Outgoing message queue--if socket full */
    struct DBuf recvQ;		/* Hold for data incoming yet to be parsed */
    long sendM;			/* Statistics: protocol messages send */
    long sendK;			/* Statistics: total k-bytes send */
    long receiveM;		/* Statistics: protocol messages received */
    long receiveK;		/* Statistics: total k-bytes received */
    u_short sendB;		/* counters to count upto 1-k lots of bytes */
    u_short receiveB;		/* sent and received. */
    long lastrecvM;		/* to check for activity --Mika */
    int priority;
    aClient *acpt;		/* listening client which we accepted from */
    Link *confs;		/* Configuration record associated */
    int authfd;			/* fd for rfc931 authentication */
    char username[USERLEN+1];	/* username here now for auth stuff */
    unsigned short port;	/* and the remote port# too :-) */
    struct hostent *hostp;
#ifdef ANTI_NICK_FLOOD
    time_t last_nick_change;
    int number_of_nick_changes;
#endif
#ifdef NO_AWAY_FLUD
    time_t alas;		/* last time of away set */
    int acount;			/* count of away settings */
#endif
    char sockhost[HOSTLEN+1];	/* This is the host name from the socket and after which the connection was accepted. */
    char passwd[PASSWDLEN+1];
    int oflag;			/* Operator Flags */
    int sockerr;		/* what was the last error returned for this socket? */
    int capabilities;		/* what this server/client supports */
    int pingval;		/* cache client class ping value here */
    int sendqlen;		/* cache client max sendq here */
};

#define	CLIENT_LOCAL_SIZE sizeof(aClient)
#define	CLIENT_REMOTE_SIZE offsetof(aClient,count)
/* statistics structures */
struct stats
{
    unsigned int is_cl;     /* number of client connections */
    unsigned int is_sv;     /* number of server connections */
    unsigned int is_ni;     /* connection but no idea who it was */
    unsigned short is_cbs;  /* bytes sent to clients */
    unsigned short is_cbr;  /* bytes received to clients */
    unsigned short is_sbs;  /* bytes sent to servers */
    unsigned short is_sbr;  /* bytes received to servers */
    unsigned long is_cks;   /* k-bytes sent to clients */
    unsigned long is_ckr;   /* k-bytes received to clients */
    unsigned long is_sks;   /* k-bytes sent to servers */
    unsigned long is_skr;   /* k-bytes received to servers */
    time_t is_cti;	    /* time spent connected by clients */
    time_t is_sti;	    /* time spent connected by servers */
    unsigned int is_ac;     /* connections accepted */
    unsigned int is_ref;    /* accepts refused */
    unsigned int is_unco;   /* unknown commands */
    unsigned int is_wrdi;   /* command going in wrong direction */
    unsigned int is_unpf;   /* unknown prefix */
    unsigned int is_empt;   /* empty message */
    unsigned int is_num;    /* numeric message */
    unsigned int is_kill;   /* number of kills generated on collisions */
    unsigned int is_fake;   /* MODE 'fakes' */
    unsigned int is_asuc;   /* successful auth requests */
    unsigned int is_abad;   /* bad auth requests */
    unsigned int is_udp;    /* packets recv'd on udp port */
    unsigned int is_loc;    /* local connections made */
#ifdef FLUD
    unsigned int is_flud;   /* users/channels flood protected */
#endif	                    /* FLUD */
};

/* mode structure for channels */
struct SMode 
{
    unsigned int mode;
    int limit;
    char key[KEYLEN+1];
};

/* Message table structure */
struct Message 
{
    char *cmd;
    int (*func) ();
    unsigned int count;			/* number of times command used */
    int parameters;
    char flags;
    
    /* bit 0 set means that this command is allowed to be used only on
     * the average of once per 2 seconds -SRB */
    
    /* I could have defined other bit maps to above instead of the next
     * two flags that I added. so sue me. -Dianora */
    
    char allow_unregistered_use;	/* flag if this command can be used if unregistered */
    char reset_idle;			/* flag if this command causes idle time to be reset */
    unsigned long bytes;
};

typedef struct msg_tree 
{
    char *final;
    struct Message *msg;
    struct msg_tree *pointers[26];
} MESSAGE_TREE;

/* Move BAN_INFO information out of the SLink struct its _only_ used
 * for bans, no use wasting the memory for it in any other type of
 * link. Keep in mind, doing this that it makes it slower as more
 * Malloc's/Free's have to be done, on the plus side bans are a smaller
 * percentage of SLink usage. Over all, the th+hybrid coding team came
 * to the conclusion it was worth the effort. - Dianora */

struct Ban 
{
    char *banstr;
    char *who;
    time_t when;
    u_char type;
    aBan *next;
};

struct Exempt
{
    char *exemptstr;
    char *who;
    time_t when;
    u_char type;
    anExempt *next;
};

/* channel member link structure, used for chanmember chains */
struct ChanLink 
{
    struct ChanLink *next;
    aClient *cptr;
    int flags;
    int bans;	/* for bquiet: number of bans against this user */
    int banexceptions;
};

/* general link structure used for chains */

struct SLink 
{
    struct SLink *next;
    union
    {
	aClient  *cptr;
	aChannel *chptr;
	aConfItem *aconf;
	aBan *banptr;
	anExempt *exemptptr;
	aWatch *wptr;
	char *cp;
    } value;
    int flags;
};

struct SLinkD
{
    struct SLinkD *next;
    struct SLinkD *prev;
    union
    {
	aClient    *cptr;
	aChannel   *chptr;
	aConfItem  *aconf;
	aBan       *banptr;
	aWatch *wptr;
	char       *cp;
    } value;
    int         flags;
};

/* channel structure */

struct Channel 
{
    struct Channel *nextch, *prevch, *hnextch;
    int hashv;		/* raw hash value */
    Mode mode;
    char topic[TOPICLEN+1];
    char topic_nick[NICKLEN+1];
    time_t topic_time;
    int users;
    chanMember *members;
    Link *invites;
    aBan *banlist;
    anExempt *exemptlist;
    ts_val channelts;
#ifdef FLUD
    time_t fludblock;
    struct fludbot *fluders;
#endif
    char chname[CHANNELLEN+1];
};

#define	TS_CURRENT	5	/* current TS protocol version */
#define	TS_MIN		3  	/* minimum supported TS protocol version */
#define	TS_DOESTS	0x20000000
#define	DoesTS(x)	((x)->tsinfo == TS_DOESTS)
/* Channel Related macros follow */


/* Channel related flags */
#define	CHFL_CHANOP		0x0001 /* */
#define	CHFL_VOICE		0x0002 /* */
#define	CHFL_DEOPPED		0x0004 /* */
#define CHFL_HALFOP		0x0008 /* */
#define CHFL_DEHALFOPPED	0x0010 /* */
#define CHFL_CHANPROT		0x0020 /* */
#define CHFL_DECHANPROT		0x0040 /* */
#define CHFL_CHANOWNER		0x0080 /* */
#define CHFL_DECHANOWNER	0x0100 /* */
#define	CHFL_BAN		0x0200 /* */
#define CHFL_EXEMPT		0x0400 /* */

/* ban mask types */
#define MTYP_FULL      0x01    /* mask is nick!user@host */
#define MTYP_USERHOST  0x02    /* mask is user@host */
#define MTYP_HOST      0x04    /* mask is host only */

/* Channel Visibility macros */
#define MODE_CHANOWNER		CHFL_CHANOWNER
#define MODE_DECHANOWNER	CHFL_DECHANOWNER
#define MODE_CHANPROT		CHFL_CHANPROT
#define MODE_DECHANPROT		CHFL_DECHANPROT
#define	MODE_CHANOP		CHFL_CHANOP
#define	MODE_DEOPPED  		CHFL_DEOPPED
#define MODE_HALFOP		CHFL_HALFOP
#define MODE_DEHALFOPPED	CHFL_DEHALFOPPED
#define	MODE_VOICE		CHFL_VOICE

#define	CMODE_L 0x000004
#define	CMODE_b 0x000008
#define CMODE_c 0x000010
#define CMODE_e 0x000020
#define	CMODE_i 0x000040
#define	CMODE_k 0x000080
#define	CMODE_l 0x000100
#define	CMODE_m 0x000200
#define	CMODE_n 0x000400
#define	CMODE_p 0x000800
#define CMODE_r 0x001000
#define	CMODE_s 0x002000
#define	CMODE_t 0x004000
#define CMODE_x 0x008000
#define CMODE_A 0x010000
#define CMODE_H 0x020000
#define CMODE_M 0x040000
#define CMODE_N 0x080000
#define CMODE_O 0x100000
#define CMODE_P 0x200000
#define CMODE_R 0x400000
#define CMODE_S 0x800000

/* mode flags which take another parameter (With PARAmeterS) */
#define	MODE_WPARAS (MODE_CHANOWNER|MODE_CHANPROT|MODE_CHANOP|MODE_HALFOP|MODE_VOICE|CMODE_b|CMODE_e|CMODE_k|CMODE_l|CMODE_L)

/*name invisible */
#define	SecretChannel(x) ((x) && ((x)->mode.mode & CMODE_s))
#define	HiddenChannel(x) ((x) && ((x)->mode.mode & CMODE_p))
/* channel visible */
#define	ShowChannel(v,c)	(PubChannel(c) || IsMember((v),(c)))
#define	PubChannel(x)		((!x) || ((x)->mode.mode & (CMODE_p | CMODE_s)) == 0)
#define IsMember(blah,chan) ((blah && blah->user && find_channel_link((blah->user)->channel, chan)) ? 1 : 0)
#define	IsChannelName(name) ((name) && (*(name) == '#' || *(name) == '&'))

/* Misc macros */
#define	BadPtr(x) (!(x) || (*(x) == '\0'))
#define	isvalid(c) (((c) >= 'A' && (c) < '~') || isdigit(c) || (c) == '-')

#define	MyConnect(x)		   ((x)->fd >= 0)
#define	MyClient(x) (MyConnect(x) && IsClient(x))
#define	MyOper(x)     (MyConnect(x) && IsOper(x))

/* String manipulation macros */

/* strncopynt --> strncpyzt to avoid confusion, sematics changed N must
 * be now the number of bytes in the array --msa */
#define	strncpyzt(x, y, N) do{(void)strncpy(x,y,N);x[N-1]='\0';}while(0)
#define	StrEq(x,y) (!strcmp((x),(y)))

/* used in SetMode() in channel.c and m_umode() in s_msg.c */
#define	MODE_NULL      0
#define	MODE_ADD       0x40000000
#define	MODE_DEL       0x20000000

/* return values for hunt_server() */
#define	HUNTED_NOSUCH	(-1)	/* if the hunted server is not found */
#define	HUNTED_ISME	0	/* if this server should execute the command */
#define	HUNTED_PASS	1	/* if message passed onwards successfully */

/* used when sending to #mask or $mask */
#define	MATCH_SERVER  1
#define	MATCH_HOST    2

/* used for async dns values */
#define	ASYNC_NONE	(-1)
#define	ASYNC_CLIENT	0
#define	ASYNC_CONNECT	1
#define	ASYNC_CONF	2
#define	ASYNC_SERVER	3

/* misc variable externs */
extern char version[128], protoctl[512], *infotext[];
extern char *generation, *creation;

/* misc defines */
#define ZIP_NEXT_BUFFER -4
#define RC4_NEXT_BUFFER -3
#define	FLUSH_BUFFER	-2
#define	UTMP		"/etc/utmp"
#define	COMMA		","

#ifdef ORATIMING
/* Timing stuff (for performance measurements): compile with
 * -DORATIMING and put a TMRESET where you want the counter of time
 * spent set to 0, a TMPRINT where you want the accumulated results,
 * and TMYES/TMNO pairs around the parts you want timed -orabidoo */
extern struct timeval tsdnow, tsdthen;
extern unsigned long tsdms;

#define TMRESET tsdms=0;
#define TMYES gettimeofday(&tsdthen, NULL);
#define TMNO gettimeofday(&tsdnow, NULL);\
             if (tsdnow.tv_sec!=tsdthen.tv_sec) \
                 tsdms+=1000000*(tsdnow.tv_sec-tsdthen.tv_sec);\
                 tsdms+=tsdnow.tv_usec; tsdms-=tsdthen.tv_usec;
#define TMPRINT sendto_umode(MODE_s, "Time spent: %ld ms", tsdms);
#else
#define TMRESET
#define TMYES
#define TMNO
#define TMPRINT
#endif

/* allow 5 minutes after server rejoins the network before allowing
 * chanops new channel */
#define MAX_SERVER_SPLIT_RECOVERY_TIME 5

#ifdef FLUD
struct fludbot 
{
    struct Client *fluder;
    int count;
    time_t first_msg, last_msg;
    struct fludbot *next;
};

#endif /* FLUD */

struct Watch 
{
    aWatch *hnext;
    time_t lasttime;
    Link *watch;
    char nick[1];
};

struct ListOptions 
{
    LOpts *next;
    Link *yeslist, *nolist;
    int starthash;
    short int showall;
    unsigned short usermin;
    int usermax;
    time_t currenttime;
    time_t chantimemin;
    time_t chantimemax;
    time_t topictimemin;
    time_t topictimemax;
};

typedef struct SearchOptions 
{
    int umodes;
    char *nick;
    char *user;
    char *host;
    char *gcos;
    char *ip;
    char *vhost;
    aChannel *channel;
    aClient *server;
    char umode_plus:1;
    char nick_plus:1;
    char user_plus:1;
    char host_plus:1;
    char gcos_plus:1;
    char ip_plus:1;
    char vhost_plus:1;
    char chan_plus:1;
    char serv_plus:1;
    char away_plus:1;
    char check_away:1;
    char check_umode:1;
    char show_chan:1;
    char search_chan:1;
    char spare:3;		/* spare space for more stuff(?) */
} SOpts;

struct t_crline
{
    char *channel;
    int type;
    aCRline *next, *prev;
};

struct t_bline
{
    char *mask;
    char *reason;
    int type;
    aBline *next;
    aBline *prev;
};

struct Gline
{
    char *set_on;
    char *set_by;	/* who set gline */
    char *username;	/* username for user@host */
    char *hostname;	/* hostname for user@host */
    time_t expire;	/* when gline expires (seconds) */
    time_t set_at;	/* when gline was set */
    char *reason;	/* reason for gline */
    aGline *next;
    aGline *prev;
};

struct Sline
{
    char *set_on;	/* current time/date */
    char *set_by;	/* who set sline */
    char *username;	/* username for user@host */
    char *hostname;	/* hostname for user@host */
    time_t expire;	/* when sline expires (seconds) */
    time_t set_at;	/* when sline was set */
    char *reason;	/* reason for sline */
    aSline *next;
    aSline *prev;
};

#define IsSendable(x) (DBufLength(&x->sendQ) < 16384)
#define DoList(x) (((x)->user) && ((x)->user->lopt))

/* internal defines for cptr->sockerr */
#define IRCERR_BUFALLOC	   -11
#define IRCERR_ZIP	   -12
#endif /* __struct_include__ */
