/* m_services.c - Because s_user.c was just crazy.
 * Copyright (C) 1990 Jarkko Oikarinen and
 * University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free softwmare; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "dh.h"
#include "h.h"

/* Externally defined stuffs */
extern int user_modes[];

/* internally defined stuffs */

/*
 * m_ns()
 */
int m_ns(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using NickServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: NickServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s NS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], NICKSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], NICKSERV);
    return 0;
}

/*
 * m_cs()
 */
int m_cs(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using ChanServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: ChanServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], CHANSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], CHANSERV);
    return 0;
}

/*
 * m_ms()
 */
int m_ms(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using MemoServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: MemoServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s MS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], MEMOSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], MEMOSERV);
    return 0;
}

/*
 * m_bs()
 */
int m_bs(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using BotServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: BotServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s BS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], BOTSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], BOTSERV);
    return 0;
}

/*
 * m_os()
 */
int m_os(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], OPERSERV);
	return 0;
    }
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s OS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], OPERSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], OPERSERV);
    return 0;
}

/*
 * m_ss()
 */
int m_ss(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using StatServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) target: StatServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s SS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], STATSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], STATSERV);
    return 0;
}

/*
 * m_hs()
 */
int m_hs(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using HostServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: HostServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s HS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], HOSTSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], HOSTSERV);
    return 0;
}

/*
 * m_rs()
 */
int m_rs(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using RootServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: RootServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], ROOTSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], ROOTSERV);
    return 0;
}

/*
 * m_gs()
 */
int m_gs(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
        
    if (check_registered_user(sptr))
        return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
        if (MyClient(sptr))
            sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
        return -1;
    } 
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
        if ((sptr->alas + MAX_MSG_TIME) < NOW)
            sptr->acount = 0;
        sptr->alas = NOW;
        sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
        sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using GlobalServ again.", 
		me.name, sptr->name, MAX_MSG_TIME);
        sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: GlobalServ", parv[0], sptr->user->username, sptr->user->host, me.name);
        sptr->since += 5;
        return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
        sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
        sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], GLOBALSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
        sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], GLOBALSERV);
    return 0;
}

/*
 * m_as()
 */
int m_as(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using AdminServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: AdminServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], ADMINSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], ADMINSERV);
    return 0;
}

/*
 * m_massserv()
 */
int m_massserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using MassServ again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: MassServ", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], MASSSERV, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], MASSSERV);
    return 0;
}

/*
 * m_agent()
 */
int m_agent(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using Agent again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: Agent", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], AGENT, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], AGENT);
    return 0;
}

/*
 * m_x()
 */
int m_x(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using X again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: X", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], X, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], X);
    return 0;
}

/*
 * m_w()
 */
int m_w(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
#ifdef NO_SERVICES_FLUD
    if (MyClient(sptr) && !IsAnOper(sptr))
    {
	if ((sptr->alas + MAX_MSG_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
    if (MyClient(sptr) && (sptr->acount > MAX_MSG_COUNT) && !IsAnOper(sptr))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Excessive services usage, wait %d seconds before using W again.",
	    me.name, sptr->name, MAX_MSG_TIME);
	sendto_umode(MODE_f, "Possible Services Flooder %s (%s@%s) on %s target: W", parv[0], sptr->user->username, sptr->user->host, me.name);
	sptr->since += 5;
	return 0;
    }
#endif
    if ((acptr = find_server(DYN_SERVER_SERVICES, NULL)))
#ifdef SERVICESHUB
	sendto_one(acptr, ":%s CS :%s", parv[0], parv[1]);
#else
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], W, DYN_SERVER_SERVICES, parv[1]);
#endif
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], W);
    return 0;
}

/*
 * m_services()
 */
int m_services(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *tmps;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((strlen(parv[1]) >= 4) && (!myncmp(parv[1], "help", 4)))
    {
	sendto_one(sptr, ":services!service@%s NOTICE %s :For ChanServ help use: /chanserv help", DYN_SERVER_SERVICES, sptr->name);
	sendto_one(sptr, ":services!service@%s NOTICE %s :For NickServ help use: /nickserv help", DYN_SERVER_SERVICES, sptr->name);
	sendto_one(sptr, ":services!service@%s NOTICE %s :For MemoServ help use: /memoserv help", DYN_SERVER_SERVICES, sptr->name);
	sendto_one(sptr, ":services!service@%s NOTICE %s :For BotServ help use: /botserv help", DYN_SERVER_SERVICES, sptr->name);
	sptr->since += 5;
	return 0;
    }
    if ((tmps = (char *) strchr(parv[1], ' ')))
    {
	for(; *tmps == ' '; tmps++); /* er.. before this for loop, the next comparison would always compare '#' with ' '.. oops. - lucas */
	if (*tmps == '#')
	    return m_cs(cptr, sptr, parc, parv);
	else
	    return m_ns(cptr, sptr, parc, parv);
    }
    return m_ns(cptr, sptr, parc, parv);
}

/*
 * m_identify()
 */
int m_identify(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char buf[BUFSIZE];
    char *myparv[parc];

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0')
    {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    (void) ircsnprintf(buf, BUFSIZE, "IDENTIFY %s", parv[1]);

    myparv[0]=parv[0];
    myparv[1]=buf;

    if (*parv[1] == '#')
      return m_cs(cptr, sptr, parc, myparv);
    else
      return m_ns(cptr, sptr, parc, myparv);
    return 0;
}

/*
 * s_svsnick()
 *  parv[0] = sender
 *  parv[1] = old nickname
 *  parv[2] = new nickname
 *  parv[3] = timestamp
 */
int m_svsnick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (!IsULine(sptr) || parc < 4 || (strlen(parv[2]) > NICKLEN))
	return 0;

    /* if we can't SVSNICK them to something because the nick is in use, KILL them */

    acptr=find_person(parv[2], NULL);

    if (acptr!=NULL)
    {
	/* send a kill out for the person instead --wd */
	sendto_serv_butone(cptr, ":%s KILL %s :%s Nickname collision due to Services enforced nickname change, your nick was overruled",
	    sptr->name, parv[1], sptr->name);
	/* now send back a kill for the nick they were presumably changed to */
	sendto_one(cptr, ":%s KILL %s :%s Nickname collision due to Services enforced nickname change, your nick was overruled",
	    sptr->name, parv[2], sptr->name);
	return 0;
    }
    if (!hunt_server(cptr, sptr, ":%s SVSNICK %s %s :%s", 1, parc, parv) != HUNTED_ISME)
    {
	if ((acptr = find_person(parv[1], NULL))!=NULL)
	{
	    acptr->umode &= ~UMODE_r;
#ifdef ANTI_NICK_FLOOD
	    acptr->last_nick_change = atoi(parv[3]);
#endif
	    sendto_common_channels(acptr, ":%s NICK :%s", parv[1], parv[2]);
	    if (IsPerson(acptr)) add_history(acptr, 1);
	    sendto_serv_butone(NULL, ":%s NICK %s :%i", parv[1], parv[2], atoi(parv[3]));
	    if (acptr->name[0]) del_from_client_hash_table(acptr->name, acptr);
	    strcpy(acptr->name, parv[2]);
	    add_to_client_hash_table(parv[2], acptr);
	}
    }
    return 0;
}

/*
 * channel_svsmode()
 *  parv[0] = sender
 *  parv[1] = channel
 *  parv[2] = modes
 *  parv[3] = nick
 *  parv[4] = nickts
 */
int channel_svsmode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel *chptr;
    aClient *acptr = NULL;
    char *m, *nick = NULL;
    char change = '+';
    ts_val nickts = 0;
    int sendmsg = 1;

    if (!(chptr = find_channel(parv[1], NULL)))
	return 0;
    if (parc >= 4)
    {
	nick = parv[3];
	if (parc > 4)
	    nickts = atol(parv[4]);
    }
    if (nick)
    {
	acptr = find_person(nick, NULL);
	if (!acptr || (nickts && acptr->tsinfo != nickts))
	    return 0;
    }

    for (m = parv[2]; *m; m++)
	switch(*m)
	{
	case '+':
	case '-':
            change = *m;
            break;
	case 'b':
            if (nick && MyClient(acptr) && change == '-')
            {
		remove_matching_bans(chptr, acptr, &me);
		sendmsg--;
            }
            break;
	case 'e':
	    if (nick && MyClient(acptr) && change == '-')
	    {
		remove_matching_exempts(chptr, acptr, &me);
		sendmsg--;
	    }
	    break;
	default:
            sendmsg++;
            break;
	}
    if (!sendmsg)
	return 0;
    if (nick)
	sendto_serv_butone(cptr, ":%s SVSMODE %s %s %s %ld", parv[0], parv[1], parv[2], nick, acptr->tsinfo);
    else
	sendto_serv_butone(cptr, ":%s SVSMODE %s %s", parv[0], parv[1], parv[2]);
    return 0;
}

/*
 * m_svsmode()
 *  parv[0] = sender
 *  parv[1] = nick
 *  parv[2] = TS (or mode, depending on svs version)
 *  parv[3] = mode (or services id if old svs version)
 *  parv[4] = optional arguement (services id)
 */
int m_svsmode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int flag, *s, what, setflags;
    char *m, *modes, *optarg;
    aClient *acptr;
    ts_val ts = 0;

    if (!IsULine(sptr) || (parc < 3))
	return 0;
    if (parv[1][0] == '#')
	return channel_svsmode(cptr, sptr, parc, parv);
    if ((parc >= 4) && ((parv[3][0] == '+') || (parv[3][0] == '-')))
    {
	ts = atol(parv[2]);
	modes = parv[3];
	optarg = (parc > 4) ? parv[4] : NULL;
    }
    else
    {
	modes = parv[2];
	optarg = (parc > 3) ? parv[3] : NULL;
    }
    if (!(acptr = find_person(parv[1], NULL)))
	return 0;
    if (ts && (ts != acptr->tsinfo))
	return 0;

    what = MODE_ADD;
    setflags = 0;

    for (s = user_modes; (flag = *s); s += 2)
	if (acptr->umode & flag)
	    setflags |= flag;
    for (m = modes; *m; m++)
	switch(*m)
	{
	case '+':
	    what = MODE_ADD;
	    break;
	case '-':
	    what = MODE_DEL;
	    break;
	case ' ':
	case '\n':
	case '\r':
	case '\t':
	    break;
	case 'd':
	    if (optarg && isdigit(*optarg))
		acptr->user->servicestamp = strtoul(optarg, NULL, 0);
	    break;
	default:
	    for (s = user_modes; (flag = *s); s += 2)
	    {
		if (*m == (char)(*(s+1)))
		{
		    if (what == MODE_ADD)
			acptr->umode |= flag;
		    else
			acptr->umode &= ~flag;
		    break;
		}
	    }
	    break;
	}
    if (optarg)
	sendto_serv_butone(cptr, ":%s SVSMODE %s %ld %s %s", parv[0], parv[1], acptr->tsinfo, modes, optarg);
    else
	sendto_serv_butone(cptr, ":%s SVSMODE %s %ld %s", parv[0], parv[1], acptr->tsinfo, modes);
    return 0;
}

/*
 * m_svschghost()
 *  parv[0] = sender prefix
 *  parv[1] = user to change hostname of
 *  parv[2] = hostname to change to
 */
int m_svschghost(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *s, *hostname;
    int legalhost = 1;

    if (!IsULine(sptr) && !IsServer(cptr))
	return 0;
    if (parc != 3)
	return 0;
    if (strlen(parv[2]) < 1)
	return 0;
    if (strlen(parv[2]) > (HOSTLEN - 1))
	return 0;

    for (s = parv[2]; *s; s++)
    {
	if (!isallowed(*s))
	{
	    legalhost = 0;
	}
    }
    if (legalhost == 0)
	return 0;

    hostname = parv[2];

    if (hostname == NULL)
	return 0;
    if ((acptr = find_person(parv[1], NULL)))
    {
	if (!HasUmodex(acptr))
	    SetUmodex(acptr);
	strncpyzt(acptr->user->virthost, hostname, HOSTLEN);
	sendto_serv_butone(cptr, ":%s SVSCHGHOST %s :%s", parv[0], parv[1], hostname);
    }
    return 0;
}

/*
 * m_svswhois()
 *  parv[0] = sender prefix
 *  parv[1] = user to change swhois of
 *  parv[2] = new swhois
 */
int m_svswhois(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (!IsServer(sptr) && !IsULine(sptr))
	return 0;
    if (parc != 3)
	return 0;
    if (strlen(parv[2]) < 1)
	return 0;
    if (strlen(parv[2]) > (HOSTLEN - 1))
	return 0;

    acptr = find_person(parv[1], (aClient *)NULL);

    if (!acptr)
	return 0;
    strncpyzt(acptr->user->swhois, parv[2], HOSTLEN);
    sendto_serv_butone(cptr, ":%s SVSWHOIS %s :%s", me.name, parv[1], parv[2]);
    return 0;
}

/*
 * m_svsjoin()
 *  parv[0] = sender
 *  parv[1] = nick to make join
 *  parv[2] = channel(s) to join
 */
int m_svsjoin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (!IsServer(sptr) && !IsULine(sptr))
	return 0;
    if (parc != 3 || !(acptr = find_person(parv[1], NULL)))
	return 0;
    if (MyClient(acptr))
    {
	parv[0] = parv[1];
	parv[1] = parv[2];
	(void)m_join(acptr, acptr, 2, parv);
    }
    else
	sendto_one(acptr, ":%s SVSJOIN %s :%s", parv[0], parv[1], parv[2]);
    return 0;
}

/*
 * m_svspart()
 *  parv[0] = sender
 *  parv[1] = nick to make part
 *  parv[2] = channel(s) to part
 */
int m_svspart(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (!IsServer(sptr) && !IsULine(sptr))
	return 0;
    if (parc != 3 || !(acptr = find_person(parv[1], NULL)))
	return 0;
    if (MyClient(acptr))
    {
	parv[0] = parv[1];
	parv[1] = parv[2];
	(void)m_part(acptr, acptr, 2, parv);
    }
    else
	sendto_one(acptr, ":%s SVSPART %s :%s", parv[0], parv[1], parv[2]);
    return 0;
}

/*
 * m_svsnoop()
 *  parv[0] = sender prefix
 *  parv[1] = server
 *  parv[2] = +/-
 */
int SVSNOOP = 0;
int m_svsnoop(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!IsServer(sptr) && !IsULine(sptr))
	return 0;
    if (parc != 3)
	return 0;
    if (hunt_server(cptr, sptr, ":%s SVSNOOP %s :%s", 1, parc, parv) == HUNTED_ISME)
    {
	if (parv[2][0] == '+')
	{
	    SVSNOOP = 1;
	    sendto_umode(MODE_s, "This server has been placed in NOOP mode");
	}
	else
	{
	    SVSNOOP = 0;
	    sendto_umode(MODE_s, "This server is no longer in NOOP mode");
	}
    }
    return 0;
}
