/*
 * VIAGRA Internet Relay Chat Daemon, src/s_confmanip.c
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#if defined( HAVE_STRING_H )
#include <string.h>
#else
/* older unices don't have strchr/strrchr .. help them out */
#include <strings.h>
#undef strchr
#define strchr index
#endif

#include <unistd.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"

/*
*
* m_removeline
* Written on July 24, 2003 by OvErRiTe
*
*/
int m_removeline (aClient * cptr, aClient * sptr, int parc, char **parv)
{
  aConfItem *aconf;
  FILE *in, *out;
  char *c1 = "CNLINE", *c2 = "ULINE", *c3 = "HLINE", *c4 = "OLINE";

  char buf[256], temppath[256];
  ircsprintf (temppath, "%s.tmp", configfile);

  if (MyClient(sptr) && !OFlagr(sptr))
  {
      sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
  }

  if (!MyClient(sptr) && !(HasUmodeT(sptr) || HasUmodeN(sptr)) && !IsULine(sptr))
  {
      sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
  }

  /* if we havent met our syntax requirments lets bitch at them */
  if ((parc < 3) || ((!mycmp(parv[1], c1) == 0) && (!mycmp(parv[1], c2) == 0) 
	&& (!mycmp(parv[1], c3) == 0) && (!mycmp(parv[1], c4) == 0)))
  {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Syntax: /REMOVELINE [ CNLINE | ULINE | HLINE ] <servername>", me.name, parv[0]);
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Syntax: /REMOVELINE [ OLINE ] <username>", me.name, parv[0]);
        return -1;
  }

  /* we have a valid syntax so lets process */
  for (aconf = conf; aconf; aconf = aconf->next)
  {
	if ((aconf->status == CONF_CONNECT_SERVER) && (mycmp (parv[2], aconf->name) == 0) 
           && (mycmp(parv[1], c1) == 0)){
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- hit cnline section", me.name, parv[0]);
	        break;
	}
	else if (aconf->status == CONF_ULINE && (mycmp (parv[2], aconf->host) == 0)
	   && (mycmp(parv[1], c2) == 0)){
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- hit uline section", me.name, parv[0]);
	        break;
        }
	else if (aconf->status == CONF_HUB && (mycmp (parv[2], aconf->name) == 0)
           && (mycmp(parv[1], c3) == 0)){
                sendto_one(sptr, ":%s NOTICE %s :*** Notice -- hit hline section", me.name, parv[0]);
                break;
        }
  }

  /* the server didnt exists so stop */
  if (!aconf)
    {
	if (mycmp(parv[1], c4) != 0){
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- REMOVELINE: Host %s not listed in %s", me.name, parv[0], parv[2], configfile);
            return -1;
	} else {
	    /* uncomment below once oline detection is written */
	    //sendto_one(sptr, ":%s NOTICE %s :*** Notice -- REMOVELINE: Host %s not listed in %s", me.name, parv[0], parv[2], configfile);
            //return -1;
	}
    }

  /* read in the file and prepare to write, then do so */
  in = fopen (configfile, "r");
  out = fopen (temppath, "w");
  while (fgets (buf, sizeof (buf), in))
    {
	/* our actions are different depending on the type of line were removing */
	if (mycmp(parv[1], c1) == 0){
           char *temp = "C";
	   char *temp2 = "N";
           if (!( (strstr (buf, parv[2])) && ((*temp == *buf) || (*temp2 == *buf)) ))
	        fputs (buf, out);
	} else if (mycmp(parv[1], c2) == 0){
           char *temp = "U";
           if (!( (strstr (buf, parv[2])) && (*temp == *buf) ))
	        fputs (buf, out);
	} else if (mycmp(parv[1], c3) == 0){
           char *temp = "H";
           if (!( (strstr (buf, parv[2])) && (*temp == *buf) ))
                fputs (buf, out);
	} else if (mycmp(parv[1], c4) == 0){ 
           char *temp = "O";
           if (!( (strstr (buf, parv[2])) && (*temp == *buf) ))
                fputs (buf, out);
        }
   }
  fclose (in);
  fclose (out);

  /* update the config file */
  if (rename (temppath, configfile) == -1)
    {
      sendto_one (sptr, ":%s NOTICE %s :*** Notice -- ERROR: rename() failed!", me.name, parv[0]);
      return -1;
    }

  /* if all went correctly we have a successful rehash */
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, ":%s NOTICE %s : *** Notice -- The requested line(s) for %s has been removed", me.name, parv[0], parv[2]);
      sendto_serv_butone (&me, ":%s GLOBOPS :%s removed configuration line(s) for %s", me.name, parv[0], parv[2]);
    }
  else
    {
      sendto_one (sptr, ":%s NOTICE %s :*** Notice -- ERROR: rehash() failed!", me.name, parv[0]);
      return -1;
    }

  return 0;
}

